__all__ = ('ThreadProfile',)

from scarletio import RichAttributeErrorBaseType

from ...utils import DISCORD_EPOCH_START

from .fields import (
    parse_flags, parse_joined_at, put_flags, put_joined_at, validate_flags, validate_joined_at
)
from .flags import ThreadProfileFlag



class ThreadProfile(RichAttributeErrorBaseType):
    """
    Represents an user's profile inside of a thread channel.
    
    Attributes
    ----------
    joined_at : `None`, `datetime`
        The date when the user joined the thread.
    flags : ``ThreadProfileFlag``
        user specific settings of the profile.
    """
    __slots__ = ('joined_at', 'flags',)
    
    
    def __new__(cls, *,  flags = ..., joined_at = ...):
        """
        Creates a new thread profile with the given parameters.
        
        Parameters
        ----------
        flags : ``ThreadProfileFlag``, `int`, Optional (Keyword only)
            user specific settings of the profile.
        joined_at : `None`, `datetime`, Optional (Keyword only)
            The date when the user joined the thread.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        # flags
        if flags is ...:
            flags = ThreadProfileFlag()
        else:
            flags = validate_flags(flags)
        
        # joined_at
        if joined_at is ...:
            joined_at = None
        else:
            joined_at = validate_joined_at(joined_at)
        
        self = object.__new__(cls)
        self.flags = flags
        self.joined_at = joined_at
        return self
    
    
    def __repr__(self):
        """Returns the thread profile's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    def __hash__(self):
        """Returns the thread profile's hash value."""
        hash_value = 0
        
        # flags
        hash_value ^= self.flags
        
        # joined_at
        joined_at = self.joined_at
        if (joined_at is not None):
            hash_value ^= hash(joined_at)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two thread profiles are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # flags
        if self.flags != other.flags:
            return False
        
        # joined_at
        if self.joined_at != other.joined_at:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a thread profile from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received thread profile data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.joined_at = parse_joined_at(data)
        self._update_attributes(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the thread profile to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        put_flags(self.flags, data, defaults)
        
        # joined_at
        if include_internals:
            put_joined_at(self.joined_at, data, defaults)
        
        return data
    
    
    def _update_attributes(self, data):
        """
        Updates the thread profile with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received thread profile data.
        """
        self.flags = parse_flags(data)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the thread profile and returns it's changed attributes in a `dict` within `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `object`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +-------------------+-------------------------------+
        | Keys              | Values                        |
        +===================+===============================+
        | flags             | ``ThreadProfileFlag``         |
        +-------------------+-------------------------------+
        """
        old_attributes = {}
        
        flags = parse_flags(data)
        if self.flags != flags:
            old_attributes['flags'] = self.flags
            self.flags = flags
        
        return old_attributes
    
    
    def copy(self):
        """
        Copies the thread profile.
        
        Returns
        -------
        new : `instance<type<self>>
        """
        new = object.__new__(type(self))
        new.flags = self.flags
        new.joined_at = self.joined_at
        return new
    
    
    def copy_with(self, *, flags = ..., joined_at = ...):
        """
        Copies the thread profile and modifies the defined the defined fields of it.
        
        Parameters
        ----------
        flags : ``ThreadProfileFlag``, `int`, Optional (Keyword only)
            user specific settings of the profile.
        joined_at : `None`, `datetime`, Optional (Keyword only)
            The date when the user joined the thread.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """

        # flags
        if flags is ...:
            flags = self.flags
        else:
            flags = validate_flags(flags)
        
        # joined_at
        if joined_at is ...:
            joined_at = self.joined_at
        else:
            joined_at = validate_joined_at(joined_at)
        
        new = object.__new__(type(self))
        new.flags = flags
        new.joined_at = joined_at
        return new
    
    
    @property
    def created_at(self):
        """
        Returns ``.joined_at`` if set, else the Discord epoch.
        
        Returns
        -------
        created_at : `datetime`
        """
        created_at = self.joined_at
        if created_at is None:
            created_at = DISCORD_EPOCH_START
        
        return created_at



__all__ = ()

from ..onboarding_prompt.fields import put_id as put_prompt_id

from .fields import (
    put_default_channel_ids, put_enabled, put_mode, put_prompts, validate_default_channel_ids,
    validate_enabled, validate_mode, validate_prompts
)


ONBOARDING_FIELD_CONVERTERS = {
    'default_channel_ids': (validate_default_channel_ids, put_default_channel_ids),
    'enabled': (validate_enabled, put_enabled),
    'mode': (validate_mode, put_mode),
    'prompts': (validate_prompts, put_prompts),
}


# https://github.com/discord/discord-api-docs/pull/6479
def flatten_emoji_data_in_onboarding_screen_prompt_options(data):
    """
    Flattens an onboarding screen's prompts' options' emojis' datas.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Onboarding prompt data.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    prompt_datas = data.get('prompts', None)
    if (prompt_datas is not None):
        for prompt_data in prompt_datas:
            option_datas = prompt_data.get('options', None)
            if (option_datas is None):
                continue
            
            for option_data in option_datas:
                try:
                    emoji_data = option_data.pop('emoji')
                except KeyError:
                    continue
                
                if emoji_data is None:
                    option_data['emoji_name'] = None
                    continue
                
                for key, value in emoji_data.items():
                    option_data['emoji_' + key] = value

    return data


def populate_prompt_ids_in_onboarding_screen_prompt_options(data, onboarding_screen_template):
    """
    Populates prompt identifiers in onboarding screen's prompts' options' emojis' datas.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Onboarding prompt data.
    
    onboarding_screen_template : `None | OnboardingScreen`
        Onboarding screen used as a template.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    prompt_datas = data.get('prompts', None)
    if (prompt_datas is not None):
        assign_ids_for = [*prompt_datas]
        
        if onboarding_screen_template is None:
            prompts = None
        else:
            prompts = onboarding_screen_template.prompts
            
        if (prompts is not None):
            for prompt in reversed(prompts):
                if not prompt.id:
                    continue
                
                prompt_data_to_match = prompt.to_data(defaults = True)
                
                for index in reversed(range(len(assign_ids_for))):
                    prompt_data = assign_ids_for[index]
                    if prompt_data != prompt_data_to_match:
                        continue
                    
                    del assign_ids_for[index]
                    put_prompt_id(prompt.id, prompt_data, True)
                    break
        
        for generated_identifier, prompt_data in enumerate(assign_ids_for, 1 << 22):
            put_prompt_id(generated_identifier, prompt_data, True)
    
    return data



__all__ = ('EmbeddedActivityUserState',)

from scarletio import RichAttributeErrorBaseType

from ...user import ZEROUSER

from .fields import (
    parse_nonce, parse_session_id, parse_user, put_nonce, put_session_id, put_user, validate_nonce,
    validate_session_id, validate_user
)


class EmbeddedActivityUserState(RichAttributeErrorBaseType):
    """
    A user's state in an embedded activity.
    
    Attributes
    ----------
    nonce : `None | str`
        Can be used to identify the state on creation.
    
    session_id : `str`
        The state's session's identifier.
    
    user : ``ClientUserBase``
        The represented user.
    """
    __slots__ = ('nonce', 'session_id', 'user')
    
    def __new__(cls, *, nonce = ..., session_id = ..., user = ...):
        """
        Creates a new embedded user state instance.
        
        Parameters
        ----------
        nonce : `None | str`, Optional (Keyword only)
            Can be used to identify the state on creation.
        
        session_id : `None | str`, Optional (Keyword only)
            The state's session's identifier.
        
        user : `None | ClientUserBase`, Optional (Keyword only)
            The represented user.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - if a parameter's value is incorrect.
        """
        # nonce
        if nonce is ...:
            nonce = None
        else:
            nonce = validate_nonce(nonce)
        
        # session_id
        if session_id is ...:
            session_id = ''
        else:
            session_id = validate_session_id(session_id)
        
        # user
        if user is ...:
            user = ZEROUSER
        else:
            user = validate_user(user)
        
        # Construct
        self = object.__new__(cls)
        self.nonce = nonce
        self.session_id = session_id
        self.user = user
        return self
    
    
    def __repr__(self):
        """Returns the embedded activity user state's representation"""
        repr_parts = ['<', type(self).__name__]
        
        # user
        repr_parts.append(' user = ')
        repr_parts.append(repr(self.user))
        
        # session_id
        repr_parts.append(', session_id = ')
        repr_parts.append(repr(self.session_id))
        
        # nonce
        nonce = self.nonce
        if (nonce is not None):
            repr_parts.append(', nonce = ')
            repr_parts.append(repr(nonce))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the embedded activity user state's hash value"""
        hash_value = 0
        
        # nonce
        nonce = self.nonce
        if (nonce is not None):
            hash_value ^= hash(nonce)
        
        # session_id
        hash_value ^= hash(self.session_id)
        
        # user
        hash_value ^= hash(self.user)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two embedded activity user state's are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # nonce
        if self.nonce != other.nonce:
            return False
        
        # session_id
        if self.session_id != other.session_id:
            return False
        
        # user
        if self.user is not other.user:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data, guild_id = 0):
        """
        Creates a new embedded activity user state from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data to create instance from.
        
        guild_id : `int` = `0`
            The respective guild's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.nonce = parse_nonce(data)
        self.session_id = parse_session_id(data)
        self.user = parse_user(data, guild_id)
        return self
    
    
    def to_data(self, *, defaults = False, guild_id = 0):
        """
        Serialises the embedded activity user state.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        guild_id : `int` = `0`, Optional (Keyword only)
            The respective guild's identifier.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_nonce(self.nonce, data, defaults)
        put_session_id(self.session_id, data, defaults)
        put_user(self.user, data, defaults, guild_id = guild_id)
        return data
    
    
    def copy(self):
        """
        Copies the embedded activity user state.
        
        Returns
        -------
        new : `instance<type<cls>>`
        """
        new = object.__new__(type(self))
        new.nonce = self.nonce
        new.session_id = self.session_id
        new.user = self.user
        return new
        
    
    def copy_with(self, *, nonce = ..., session_id = ..., user = ...):
        """
        Copies the embedded activity user state with the given fields.
        
        Parameters
        ----------
        nonce : `None | str`, Optional (Keyword only)
            Can be used to identify the state on creation.
        
        session_id : `None | str`, Optional (Keyword only)
            The state's session's identifier.
        
        user : `None | ClientUserBase`, Optional (Keyword only)
            The represented user.
        
        Returns
        -------
        new : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - if a parameter's value is incorrect.
        """
        # nonce
        if nonce is ...:
            nonce = self.nonce
        else:
            nonce = validate_nonce(nonce)
        
        # session_id
        if session_id is ...:
            session_id = self.session_id
        else:
            session_id = validate_session_id(session_id)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
        
        # Construct
        new = object.__new__(type(self))
        new.nonce = nonce
        new.session_id = session_id
        new.user = user
        return new
    
    
    @property
    def user_id(self):
        """
        Returns the represented user's identifier.
        
        Returns
        -------
        user_id : `int`
        """
        return self.user.id



__all__ = ()

from ...color import Color

from .preinstanced import ActivityType


ACTIVITY_CUSTOM_ID_DEFAULT = 'UNKNOWN'

ACTIVITY_CUSTOM_IDS = {
    ActivityType.spotify: 'spotify:1',
    ActivityType.custom: 'custom',
}

ACTIVITY_DEFAULT_NAME = 'Unknown'


ACTIVITY_COLOR_GAME = Color(0x7289da)
ACTIVITY_COLOR_STREAM = Color(0x593695)
ACTIVITY_COLOR_SPOTIFY = Color(0x1db954)
ACTIVITY_COLOR_NONE = Color()



__all__ = ()

import reprlib
from base64 import b64encode

from scarletio import Compound

from ...http import DiscordApiClient
from ...payload_building import build_create_payload, build_edit_payload
from ...soundboard import SoundboardSound
from ...soundboard.soundboard_sound.utils import SOUNDBOARD_SOUND_FIELD_CONVERTERS

from ..request_helpers import (
    get_guild_id, get_soundboard_sound_and_guild_id_and_id, get_soundboard_sound_guild_id_and_id
)


class ClientCompoundSoundBoardEndpoints(Compound):
    
    api : DiscordApiClient
    
    async def soundboard_sound_get_all_default(self):
        """
        Requests the default soundboard sounds.
        
        This method is a coroutine.
        
        Returns
        -------
        sounds : `list` of ``SoundboardSound``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        sound_datas = await self.api.soundboard_sound_get_all_default()
        return [SoundboardSound.from_data(sound_data) for sound_data in sound_datas]
    
    
    async def soundboard_sound_create(
        self, guild, sound, soundboard_sound_template = None, *, reason = None, **keyword_parameters
    ):
        """
        Creates a soundboard sound at the given guild.
        
        This function is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild where the soundboard sound will be created.
        
        sound : `bytes-like`
            The sound to add. Expected to be in `mp3` format.
        
        soundboard_sound_template : `None`, ``SoundboardSound`` = `None`, Optional
            soundboard sound entity to use as a template.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to create the soundboard_sound with.
        
        Other Parameters
        ----------------
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            Emoji assigned to the sound.
        
        name : `str`, Optional (Keyword only)
            The name of the sound.
        
        volume : `float`, Optional (Keyword only)
            The volume of the sound to play as.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        if not isinstance(sound, (bytes, bytearray, memoryview)):
            raise TypeError(
                f'`sound_data` can be `bytes-like`, got {sound.__class__.__name__}; {reprlib.repr(sound)}.'
            )
        
        data = build_create_payload(soundboard_sound_template, SOUNDBOARD_SOUND_FIELD_CONVERTERS, keyword_parameters)
        data['sound'] = 'data:audio/mp3;base64,' + b64encode(sound).decode('ascii')
        
        sound_data = await self.api.soundboard_sound_create(guild_id, data, reason)
        return SoundboardSound.from_data(sound_data)
    
    
    async def soundboard_sound_edit(
        self, soundboard_sound, soundboard_sound_template = None, *, reason = None, **keyword_parameters
    ):
        """
        Edits the soundboard sound with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        soundboard_sound : ``SoundboardSound``, `tuple` (`int`, `int`)
            The soundboard sound to edit.
        
        soundboard_sound_template : `None`, ``SoundboardSound`` = `None`, Optional
            Soundboard sound entity to use as a template.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to create the soundboard sound with.
        
        Other Parameters
        ----------------
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            Emoji assigned to the sound.
        
        name : `str`, Optional (Keyword only)
            The name of the sound.
        
        volume : `float`, Optional (Keyword only)
            The volume of the sound to play as.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        soundboard_sound, guild_id, soundboard_sound_id = get_soundboard_sound_and_guild_id_and_id(soundboard_sound)
        
        data = build_edit_payload(
            soundboard_sound, soundboard_sound_template, SOUNDBOARD_SOUND_FIELD_CONVERTERS, keyword_parameters
        )
        
        if data:
            await self.api.soundboard_sound_edit(guild_id, soundboard_sound_id, data, reason)
    
    
    async def soundboard_sound_delete(self, soundboard_sound, *, reason = None):
        """
        Deletes the given soundboard_sound.
        
        This method is a coroutine.
        
        Parameters
        ----------
        soundboard_sound : ``SoundboardSound``, `tuple` (`int`, `int`)
            The soundboard sound to delete
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `soundboard_sound` was not given neither as ``SoundboardSound`` nor as `tuple` of (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, soundboard_sound_id = get_soundboard_sound_guild_id_and_id(soundboard_sound)
        await self.api.soundboard_sound_delete(guild_id, soundboard_sound_id, reason)
    
    # Not a thing (yet?)
    '''
    async def soundboard_sound_get(self, soundboard_sound, *, force_update = False):
        """
        Requests the specified soundboard sound.
        
        This method is a coroutine.
        
        Parameters
        ----------
        soundboard_sound : ``SoundboardSound``, `tuple` (`int`, `int`)
            The soundboard_sound to get, or a `guild-id`, `soundboard-sound-id` tuple representing it.
        
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the scheduled event should be requested even if it supposed to be up to date.
        
        Returns
        -------
        soundboard_sound : ``SoundboardSound``
        
        Raises
        ------
        TypeError
            - If `soundboard_sound`'s type is neither ``SoundboardSound``, nor `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        soundboard_sound, guild_id, soundboard_sound_id = get_soundboard_sound_and_guild_id_and_id(soundboard_sound)
        
        soundboard_sound_data = await self.api.soundboard_sound_get(guild_id, soundboard_sound_id)
        
        if (soundboard_sound is None) or force_update or soundboard_sound.partial:
            soundboard_sound, is_created = SoundboardSound.from_data_is_created(soundboard_sound_data)
            if not is_created:
                soundboard_sound._set_attributes(soundboard_sound_data, False)
        
        else:
            soundboard_sound = SoundboardSound.from_data(soundboard_sound_data)
        
        return soundboard_sound
    '''



__all__ = ('Resolved',)

from scarletio import RichAttributeErrorBaseType, export

from .fields import (
    parse_attachments, parse_channels, parse_messages, parse_roles, parse_users, put_attachments,
    put_channels, put_messages, put_roles, put_users, validate_attachments, validate_channels,
    validate_messages, validate_roles, validate_users
)


@export
class Resolved(RichAttributeErrorBaseType):
    """
    Contains the resolved entities by an interaction.
    
    Attributes
    ----------
    attachments : `None`, `dict` of (`int`, `Attachment``) items
        Resolved received attachments stored by their identifier as keys if any.
    
    channels : `None`, `dict` of (`int`, ``Channel``) items
        Resolved received channels stored by their identifier as keys if any.
    
    roles : `None`, `dict` of (`int`, ``Role``) items
        Resolved received roles stored by their identifier as keys if any.
    
    messages : `None`, `dict` of (`int`, ``Message``) items
        Resolved received messages stored by their identifier as keys if any.
    
    users : `None`, `dict` of (`int`, ``ClientUserBase``) items
        Resolved received users stored by their identifier as keys if any.
    """
    __slots__ = ('attachments', 'channels', 'roles', 'messages', 'users')
    
    def __new__(cls, *, attachments = None, channels = None, roles = None, messages = None, users = None):
        """
        Creates a new resolved object.
        
        Parameters
        ----------
        attachments : `None`, `dict` of (`int`, `Attachment``) items = `None`, Optional (Keyword only)
            Resolved received attachments stored by their identifier as keys if any.
        
        channels : `None`, `dict` of (`int`, ``Channel``) items = `None`, Optional (Keyword only)
            Resolved received channels stored by their identifier as keys if any.
        
        roles : `None`, `dict` of (`int`, ``Role``) items = `None`, Optional (Keyword only)
            Resolved received roles stored by their identifier as keys if any.
        
        messages : `None`, `dict` of (`int`, ``Message``) items = `None`, Optional (Keyword only)
            Resolved received messages stored by their identifier as keys if any.
        
        users : `None`, `dict` of (`int`, ``ClientUserBase``) items = `None`, Optional (Keyword only)
            Resolved received users stored by their identifier as keys if any.
        
        Raises
        ------
        TypeError
            - Parameter of incorrect type given.
        ValueError
            - Parameter of incorrect value given.
        """
        attachments = validate_attachments(attachments)
        channels = validate_channels(channels)
        roles = validate_roles(roles)
        messages = validate_messages(messages)
        users = validate_users(users)
        
        # Construct
        self = object.__new__(cls)
        self.attachments = attachments
        self.channels = channels
        self.roles = roles
        self.messages = messages
        self.users = users
        return self
    
    
    @classmethod
    def from_data(cls, data, guild_id = 0):
        """
        Creates a new resolved instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Resolved data.
        
        guild_id : `int` = `0`, Optional (Keyword only)
            The respective guild's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.attachments = parse_attachments(data)
        self.channels = parse_channels(data, guild_id)
        self.roles = parse_roles(data, guild_id)
        self.messages = parse_messages(data)
        self.users = parse_users(data, guild_id)
        return self
    
    
    def to_data(self, *, defaults = False, guild_id = 0):
        """
        Converts the resolved instance into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included as well.
        
        guild_id : `int` = `0`, Optional (Keyword only)
            The respective guild's identifier to use for handing user guild profiles.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_attachments(self.attachments, data, defaults)
        put_channels(self.channels, data, defaults)
        put_roles(self.roles, data, defaults)
        put_messages(self.messages, data, defaults)
        put_users(self.users, data, defaults, guild_id = guild_id)
        return data
    
    
    def copy(self):
        """
        Copies the resolved instance.
        
        Returns
        -------
        new : `instance<type<cls>``
        """
        # attachments
        attachments = self.attachments
        if (attachments is not None):
            attachments = attachments.copy()
        
        # channels
        channels = self.channels
        if (channels is not None):
            channels = channels.copy()
        
        # roles
        roles = self.roles
        if (roles is not None):
            roles = roles.copy()
        
        # messages
        messages = self.messages
        if (messages is not None):
            messages = messages.copy()
        
        # users
        users = self.users
        if (users is not None):
            users = users.copy()
        
        # Construct
        new = object.__new__(type(self))
        new.attachments = attachments
        new.channels = channels
        new.roles = roles
        new.messages = messages
        new.users = users
        return new
    
    
    def copy_with(self, *, attachments = ..., channels = ..., roles = ..., messages = ..., users = ...):
        """
        Copies the resolved with the specified fields.
        
        Parameters
        ----------
        attachments : `None`, `dict` of (`int`, `Attachment``) items, Optional (Keyword only)
            Resolved received attachments stored by their identifier as keys if any.
        
        channels : `None`, `dict` of (`int`, ``Channel``) items, Optional (Keyword only)
            Resolved received channels stored by their identifier as keys if any.
        
        roles : `None`, `dict` of (`int`, ``Role``) items, Optional (Keyword only)
            Resolved received roles stored by their identifier as keys if any.
        
        messages : `None`, `dict` of (`int`, ``Message``) items, Optional (Keyword only)
            Resolved received messages stored by their identifier as keys if any.
        
        users : `None`, `dict` of (`int`, ``ClientUserBase``) items, Optional (Keyword only)
            Resolved received users stored by their identifier as keys if any.
        
        Raises
        ------
        TypeError
            - Parameter of incorrect type given.
        ValueError
            - Parameter of incorrect value given.
        """
        # attachments
        if attachments is ...:
            attachments = self.attachments
            if (attachments is not None):
                attachments = attachments.copy()
        else:
            attachments = validate_attachments(attachments)
        
        # channels
        if channels is ...:
            channels = self.channels
            if (channels is not None):
                channels = channels.copy()
        else:
            channels = validate_channels(channels)
        
        # roles
        if roles is ...:
            roles = self.roles
            if (roles is not None):
                roles = roles.copy()
        else:
            roles = validate_roles(roles)
        
        # messages
        if messages is ...:
            messages = self.messages
            if (messages is not None):
                messages = messages.copy()
        else:
            messages = validate_messages(messages)
        
        # users
        if users is ...:
            users = self.users
            if (users is not None):
                users = users.copy()
        else:
            users = validate_users(users)
        
        # Construct
        new = object.__new__(type(self))
        new.attachments = attachments
        new.channels = channels
        new.roles = roles
        new.messages = messages
        new.users = users
        return new
    
    
    def __repr__(self):
        """Returns the resolved's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        field_added = False
        
        for field_name, field_value in zip(
            ('attachments', 'channels', 'roles', 'messages', 'users'), self._iter_entity_containers(),
        ):
            if (field_value is not None):
                if field_added:
                    repr_parts.append(',')
                else:
                    field_added = True
                
                repr_parts.append(' ')
                repr_parts.append(field_name)
                repr_parts.append(' = ')
                repr_parts.append(repr(field_value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two resolved are equal."""
        if type(other) is not type(self):
            return NotImplemented
        
        for self_container, other_container in zip(self._iter_entity_containers(), other._iter_entity_containers()):
            if self_container != other_container:
                return False
        
        return True
    
    
    def __hash__(self):
        """Returns the resolved's hash value."""
        hash_value = 0
        
        shift = 0
        
        for field_value in self._iter_entity_containers():
            if (field_value is not None):
                hash_value ^= len(field_value) << shift
                
                for entity_id in field_value.keys():
                    hash_value ^= entity_id
            
            shift += 4
        
        return hash_value
    
    
    def _iter_entity_containers(self):
        """
        Iterates over the entity containers of the resolved instance.
        
        This method is an iterable generator.
        
        Yields
        ------
        container : `None`, `dict` of (`str`, `object`) items
        """
        yield self.attachments
        yield self.channels
        yield self.roles
        yield self.messages
        yield self.users
    
    # Extra utility
    
    def resolve_attachment(self, attachment_id):
        """
        Tries to resolve an attachment by the given identifier.
        
        Parameters
        ----------
        attachment_id : `int`
            Attachment identifier.
        
        Returns
        -------
        attachment : `None`, ``Attachment``
        """
        attachments = self.attachments
        if (attachments is not None):
            return attachments.get(attachment_id, None)
    
    
    def resolve_channel(self, channel_id):
        """
        Tries to resolve an channel by the given identifier.
        
        Parameters
        ----------
        channel_id : `int`
            Channel identifier.
        
        Returns
        -------
        channel : `None`, ``Channel``
        """
        channels = self.channels
        if (channels is not None):
            return channels.get(channel_id, None)
    
    
    def resolve_role(self, role_id):
        """
        Tries to resolve an role by the given identifier.
        
        Parameters
        ----------
        role_id : `int`
            Role identifier.
        
        Returns
        -------
        role : `None`, ``Role``
        """
        roles = self.roles
        if (roles is not None):
            return roles.get(role_id, None)
    
    
    def resolve_message(self, message_id):
        """
        Tries to resolve an message by the given identifier.
        
        Parameters
        ----------
        message_id : `int`
            Message identifier.
        
        Returns
        -------
        message : `None`, ``Message``
        """
        messages = self.messages
        if (messages is not None):
            return messages.get(message_id, None)
    
    
    def resolve_user(self, user_id):
        """
        Tries to resolve an user by the given identifier.
        
        Parameters
        ----------
        user_id : `int`
            User identifier.
        
        Returns
        -------
        user : `None`, ``ClientUserBase``
        """
        users = self.users
        if (users is not None):
            return users.get(user_id, None)
    
    
    def resolve_mentionable(self, mentionable_id):
        """
        Tries to resolve a mentionable entity (user / role).
        
        Parameters
        ----------
        mentionable_id : `int`
            Entity identifier.
        
        Returns
        -------
        entity : `None`, ``Role``, ``ClientUserBase``
        """
        users = self.users
        if (users is not None):
            try:
                return users[mentionable_id]
            except KeyError:
                pass
        
        roles = self.roles
        if (roles is not None):
            try:
                return roles[mentionable_id]
            except KeyError:
                pass
    
    
    def resolve_entity(self, entity_id):
        """
        Tries to resolve any entity with the given identifier.
        
        Parameters
        ----------
        entity_id : `int`
            Entity identifier.
        
        Returns
        -------
        entity : `None` ``Attachment``, ``Channel``, ``ClientUserBase``, ``Role``, ``Message``
        """
        for container in self._iter_entity_containers():
            if (container is not None):
                try:
                    return container[entity_id]
                except KeyError:
                    pass



__all__ = ('AutoModerationActionMetadataBlock',)

from scarletio import copy_docs

from .base import AutoModerationActionMetadataBase
from .fields import parse_custom_message, put_custom_message, validate_custom_message


class AutoModerationActionMetadataBlock(AutoModerationActionMetadataBase):
    """
    Block action metadata of an auto moderation action.
    
    Attributes
    ----------
    custom_message : `int`
        A custom message that can be used as an explanation if a message is blocked.
    """
    __slots__ = ('custom_message',)
    
    def __new__(cls, custom_message = None):
        """
        Creates a new timeout action metadata for ``AutoModerationAction``-s.
        
        Parameters
        ----------
        custom_message : `None`, `str` = `None`, Optional
            A custom message that can be used as an explanation if a message is blocked.
        
        Raises
        ------
        TypeError
            - If `custom_message` type is incorrect.
        """
        custom_message = validate_custom_message(custom_message)
        
        self = object.__new__(cls)
        self.custom_message = custom_message
        return self
    
    
    @copy_docs(AutoModerationActionMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        custom_message = self.custom_message
        if (custom_message is not None):
            repr_parts.append(' custom_message = ')
            repr_parts.append(repr(custom_message))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationActionMetadataBase.from_data)
    def from_data(cls, data):
        custom_message = parse_custom_message(data)
        
        self = object.__new__(cls)
        self.custom_message = custom_message
        return self
    
    
    @copy_docs(AutoModerationActionMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_custom_message(self.custom_message, data, defaults)
        return data
    
    
    @copy_docs(AutoModerationActionMetadataBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.custom_message != other.custom_message:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationActionMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # custom_message
        custom_message = self.custom_message
        if (custom_message is not None):
            hash_value ^= hash(custom_message)
        
        return hash_value
    
    
    @copy_docs(AutoModerationActionMetadataBase.copy)
    def copy(self):
        new = AutoModerationActionMetadataBase.copy(self)
        
        # custom_message
        new.custom_message = self.custom_message
        
        return new
    
    
    def copy_with(self, *, custom_message = ...):
        """
        Copies the action metadata and modifies it's attributes by the given values.
        
        Parameters
        ----------
        custom_message : `None`, `str`, Optional (Keyword only)
            A custom message that can be used as an explanation if a message is blocked.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        if custom_message is ...:
            custom_message = self.custom_message
        else:
            custom_message = validate_custom_message(custom_message)
        
        new = AutoModerationActionMetadataBase.copy(self)
        new.custom_message = custom_message
        return new



from .application_command import *
from .auto_moderation_rule import *
from .channel import *
from .channel_permission_overwrite import *
from .emoji import *
from .guild import *
from .integration import *
from .invite import *
from .onboarding_prompt import *
from .onboarding_screen import *
from .role import *
from .scheduled_event import *
from .soundboard_sound import *
from .stage import *
from .sticker import *
from .user import *
from .webhook import *


__all__ = (
    *application_command.__all__,
    *auto_moderation_rule.__all__,
    *channel.__all__,
    *channel_permission_overwrite.__all__,
    *emoji.__all__,
    *guild.__all__,
    *integration.__all__,
    *invite.__all__,
    *onboarding_prompt.__all__,
    *onboarding_screen.__all__,
    *role.__all__,
    *scheduled_event.__all__,
    *soundboard_sound.__all__,
    *stage.__all__,
    *sticker.__all__,
    *user.__all__,
    *webhook.__all__,
)



__all__ = ()

from scarletio import from_json

from ...discord.application_command.helpers import maybe_locale


def validate_translation_table(translation_table):
    """
    Validates the given translation table returning a new one.
    
    Parameters
    ----------
    translation_table : `None`, `str`, `dict` of ((``Locale``, `str`), (`None`,
            `dict` of (`str`, (`None`, `str`)) items)) items
        Translation table for the commands of the slasher.
    
    Returns
    -------
    translation_table : `None`, `dict` of (``Locale``, `dict` of (`str`, `str`) items) items
    
    Raises
    ------
    FileNotFoundError
        - If `translation_table` is a string, but not a file.
    TypeError
        - If `translation_table`'s structure is incorrect.
    """
    if translation_table is None:
        return None
    
    if isinstance(translation_table, str):
        return _validate_translation_table_from_file(translation_table)
    
    if isinstance(translation_table, dict):
        return _validate_translation_table_from_dict(translation_table)
    
    raise TypeError(
        f'`translation_table` can be either `None`, `str`, `dict`, got '
        f'{translation_table.__class__.__name__}, {translation_table!r}'
    )


def _validate_translation_table_from_file(file_name):
    """
    Validates the given translation table opening the given `file_name`.
    
    Parameters
    ----------
    file_name : `str`
        The file's name to open. Can be absolute and relative path too.
    
    Returns
    -------
    translation_table : `None`, `dict` of (``Locale``, `dict` of (`str`, `str`) items) items
    
    Raises
    ------
    FileNotFoundError
        - If the file is not found.
    TypeError
        - If the loaded data structure is incorrect.
    """
    with open(file_name, 'r') as file:
        content = file.read()
    
    # empty file? No problem!
    if not content:
        return None
    
    try:
        json_data = from_json(content)
    finally:
        # Unallocate content if exception occurs
        content = None
    
    # No json? No problem!
    if json_data is None:
        return None
    
    if isinstance(json_data, dict):
        return _validate_translation_table_from_dict(json_data)
    
    raise TypeError(
        f'`translation_table` file can contain either `None` or `dict`, got '
        f'{json_data.__class__.__name__}; {json_data!r}.'
    )


def _validate_translation_table_from_dict(translation_table):
    """
    Validates the given translation table dictionary returning a new one.
    
    Parameters
    ----------
    translation_table : `dict` of ((``Locale``, `str`), `dict` of (`str`, (`None`, `str`)) items) items
        Translation table to validate
    
    Returns
    -------
    validated_translation_table : `None`, `dict` of (``Locale``, (`None`, `dict` of (`str`, `str`) items)) items
    
    Raises
    ------
    TypeError
        - If `translation_table`'s structure is incorrect.
    """
    validated_translation_table = None
    
    for locale, relations in translation_table.items():
        locale = maybe_locale(locale)
        relations = _validate_translation_table_relations(relations)
        
        if not relations:
            continue
        
        if validated_translation_table is None:
            validated_translation_table = {}
        
        validated_translation_table[locale] = relations
    
    return validated_translation_table


def _validate_translation_table_relations(relations):
    """
    Validates the given translation table's relations.
    
    Parameters
    ----------
    relations : `None`, `dict` of (`str`, (`None`, `str`)) items
        Relations to validate.
    
    Returns
    -------
    validated_relations : `None`, `dict` of (`str`, `str`) items
    
    Raises
    ------
    TypeError
        - If `relations`'s type or structure is incorrect.
    """
    if relations is None:
        return None
    
    if isinstance(relations, dict):
        return _validate_translation_table_relations_dict(relations)
    
    raise TypeError(
        f'`translation_table`\'s keys (also known as relations) can be `None`, `dict`, got '
        f'{relations.__class__.__name__}; {relations!r}.'
    )


def _validate_translation_table_relations_dict(relations):
    """
    
    Parameters
    ----------
    relations : `dict` of (`str`, (`None`, `str`)) items
        Relations to validate.
    
    Returns
    -------
    validated_relations : `None`, `dict` of (`str`, `str`) items

    Raises
    ------
    TypeError
        - If an item's structure is incorrect.
    """
    validated_relations = None
    
    for key, value in relations.items():
        if not isinstance(key, str):
            raise TypeError(
                f'`relation` keys can be `str`, got {key.__class__.__name__}; {key!r}.'
            )
        
        if value is None:
            continue
        
        if not isinstance(value, str):
            raise TypeError(
                f'`relation` values can be `str`, got {value.__class__.__name__}; {value!r}.'
            )
        
        if (not key) or (not value):
            continue
        
        if validated_relations is None:
            validated_relations = {}
        
        validated_relations[key] = value
        continue
    
    return validated_relations



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, force_string_parser_factory, negated_bool_parser_factory,
    nullable_date_time_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_putter_factory, force_bool_putter_factory, force_string_putter_factory,
    negated_bool_optional_putter_factory, nullable_date_time_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, force_string_validator_factory,
    nullable_date_time_validator_factory
)
from ..user import ClientUserBase, User, ZEROUSER

# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id = entity_id_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', NotImplemented, include = 'Channel')

# deaf

parse_deaf = bool_parser_factory('deaf', False)
put_deaf = force_bool_putter_factory('deaf')
validate_deaf = bool_validator_factory('deaf', False)

# guild_id

validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# mute

parse_mute = bool_parser_factory('mute', False)
put_mute = force_bool_putter_factory('mute')
validate_mute = bool_validator_factory('mute', False)

# requested_to_speak_at

parse_requested_to_speak_at = nullable_date_time_parser_factory('request_to_speak_timestamp')
put_requested_to_speak_at = nullable_date_time_optional_putter_factory('request_to_speak_timestamp')
validate_requested_to_speak_at = nullable_date_time_validator_factory('requested_to_speak_at')

# self_deaf

parse_self_deaf = bool_parser_factory('self_deaf', False)
put_self_deaf = force_bool_putter_factory('self_deaf')
validate_self_deaf = bool_validator_factory('self_deaf', False)

# self_mute

parse_self_mute = bool_parser_factory('self_mute', False)
put_self_mute = force_bool_putter_factory('self_mute')
validate_self_mute = bool_validator_factory('self_mute', False)

# self_stream

parse_self_stream = bool_parser_factory('self_stream', False)
put_self_stream = bool_optional_putter_factory('self_stream', False)
validate_self_stream = bool_validator_factory('self_stream', False)

# self_video

parse_self_video = bool_parser_factory('self_video', False)
put_self_video = force_bool_putter_factory('self_video')
validate_self_video = bool_validator_factory('self_video', False)

# session_id

parse_session_id = force_string_parser_factory('session_id')
put_session_id = force_string_putter_factory('session_id')
validate_session_id = force_string_validator_factory('session_id', 0, 1024)

# speaker

parse_speaker = negated_bool_parser_factory('suppress', False)
put_speaker = negated_bool_optional_putter_factory('suppress', False)
validate_speaker = bool_validator_factory('speaker', False)

# user_id

parse_user_id = entity_id_parser_factory('user_id')
put_user_id = entity_id_putter_factory('user_id')
validate_user_id = entity_id_validator_factory('user_id', ClientUserBase)

# user


def parse_user(data, guild_id = 0):
    """
    Parses out the voice state's user from the given data.
    
    Returns `None` if user could not be parsed.
        
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Interaction event data.
    guild_id : `int` = `0`, Optional (Keyword only)
        The respective guild's identifier.
    
    Returns
    -------
    user : `None | ClientUserBase`
    """
    # user
    try:
        guild_profile_data = data['member']
    except KeyError:
        user_data = data.get('user', None)
        if user_data is None:
            return None
        
        guild_profile_data = None
    else:
        user_data = guild_profile_data.get('user', None)
        if user_data is None:
            return None
    
    return User.from_data(user_data, guild_profile_data, guild_id)



__all__ = ()

from collections import deque

from scarletio import Future, LOOP_TIME, RichAttributeErrorBaseType

from ..core import KOKORO

from .constants import GATEWAY_RATE_LIMIT_LIMIT, GATEWAY_RATE_LIMIT_RESET


class GatewayRateLimiter(RichAttributeErrorBaseType):
    """
    Burst rate limit handler for gateways, what operates on the clients' loop only.
    
    Attributes
    ----------
    queue : `deque` of ``Future``
        The queue of the rate limit handler. It is filled up with futures, if the handler's limit is exhausted.
        These futures are removed and their result is set, when the limit is reset.
    remaining : `int`
        The amount of requests which can be done, before the limit is exhausted.
    resets_at : `float`
        When the rate limit of the respective gateway will be reset.
    wake_up_handle : `None`  or ``TimerHandle``
        A handler what will reset the limiter's limit and ensure it's queue if needed.
    """
    __slots__ = ('queue', 'remaining', 'resets_at', 'wake_up_handle')
    
    def __new__(cls):
        """
        Creates a gateway rate limiter.
        """
        self = object.__new__(cls)
        self.remaining = GATEWAY_RATE_LIMIT_LIMIT
        self.queue = deque()
        self.wake_up_handle = None
        self.resets_at = 0.0
        return self
    
    
    def __repr__(self):
        """Returns the gateway rate limiter's representation."""
        repr_parts = ['<', type(self).__name__,]
        
        resets_at = self.resets_at
        if resets_at <= LOOP_TIME():
            remaining = GATEWAY_RATE_LIMIT_LIMIT
        else:
            repr_parts.append(' resets_at = ')
            repr_parts.append(format(resets_at, '.2f'))
            repr_parts.append(' (monotonic),')
            
            remaining = self.remaining
        
        repr_parts.append(', remaining = ')
        repr_parts.append(repr(remaining))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __iter__(self):
        """
        Awaits the rate limit handler.
        
        This method is a generator. Should be used with `await` expression.
        
        Returns
        -------
        success : `bool`
            Whether the action can be executed.
            Returns `False` if the gateway was closed and no action should be executed.
        """
        now = LOOP_TIME()
        if now >= self.resets_at:
            self.resets_at = now + GATEWAY_RATE_LIMIT_RESET
            remaining = GATEWAY_RATE_LIMIT_LIMIT
        else:
            remaining = self.remaining
        
        if remaining:
            self.remaining = remaining - 1
            return True
        
        if self.wake_up_handle is None:
            self.wake_up_handle = KOKORO.call_at(self.resets_at, type(self).wake_up, self)
        
        future = Future(KOKORO)
        self.queue.append(future)
        return (yield from future)
    
    
    __await__ = __iter__
    
    
    def wake_up(self):
        """
        Wake ups the waiting futures of the gateway rate limiter.
        """
        queue = self.queue
        remaining = GATEWAY_RATE_LIMIT_LIMIT
        if not queue:
            wake_up_handle = None
        
        else:
            while True:
                if not queue:
                    wake_up_handle = None
                    break
                
                if not remaining:
                    self.resets_at = resets_at = LOOP_TIME() + GATEWAY_RATE_LIMIT_RESET
                    wake_up_handle = KOKORO.call_at(resets_at + GATEWAY_RATE_LIMIT_RESET, type(self).wake_up, self)
                    break
                
                # if the future is already done (expectedly cancelled) `.set_result_if_pending` returns `0`.
                # At that case we do not want to decrease `remaining`.
                if queue.popleft().set_result_if_pending(True):
                    remaining -= 1
        
        self.wake_up_handle = wake_up_handle
        self.remaining = remaining
    
    
    def cancel(self):
        """
        Cancels the gateway rate limiter's queue and it's `.wake_up_handle` if set.
        """
        queue = self.queue
        while queue:
            queue.popleft().set_result_if_pending(False)
        
        wake_up_handle = self.wake_up_handle
        if (wake_up_handle is not None):
            self.wake_up_handle = None
            wake_up_handle.cancel()



from .embedded_activity_configuration import *
from .fields import *
from .preinstanced import *


__all__ = (
    *embedded_activity_configuration.__all__,
    *fields.__all__,
    *preinstanced.__all__,
)


# Structure as in 2023-11-28:
#
#
# tp EmbeddedActivityConfiguration : Object
#     activity_preview_video_asset_id : null | String # 0-9 # rename to preview_video_asset_id
#     supported_platforms : list<PlatformType> # we can ignore this, we already got client_platform_config
#     default_orientation_lock_state : OrientationLockState
#     tablet_default_orientation_lock_state : OrientationLockState # rename to default_tablet_orientation_lock_state
#     requires_age_gate : bool # rename to age_gated
#     premium_tier_requirement : null # probably not used anymore
#     free_period_starts_at : null | DateTime # probably not used anymore
#     free_period_ends_at : null | DateTime # probably not used anymore
#     client_platform_config : null | Dictionary<PlatformType, ClientPlatformConfiguration> # rename to client_platform_configurations
#     shelf_rank : Integer # rename to position
#     has_csp_exception : Boolean # always false | csp = Content Security Policy | if true -> developers may see your ip # rename to content_security_policy_exceptions_exists
#     instance_mode : Enum<Integer<0>> # no results -> not yet used perhaps?
#
#
# tp ClientPlatformConfiguration : Object
#     label_type : LabelType
#     label_until : null | DateTime # rename to labelled_until
#     release_phase : ReleasePhase # Not used
#
#
# tp ReleasePhase : Enum<String>
#     GLOBAL_LAUNCH = "global_launch"
#
#     fn __default__(type) -> type.GLOBAL_LAUNCH
#
#
# tp LabelType : Enum<Integer>
#     NONE = 0
#     NEW = 1
#     UPDATED = 2
#
#
# tp OrientationLockState : Enum<Integer>
#     NONE = 0
#     UNLOCKED = 1
#     PORTRAIT = 2
#     LANDSCAPE = 3
#
#
# tp PlatformType : Enum<String>
#     ANDROID = "android"
#     IOS = "ios"
#     WEB = "web"
#
#     fn __default__(type) -> type.WEB



__all__ = ()

from ...integration import IntegrationExpireBehavior, IntegrationType
from ...integration.integration.fields import validate_name, validate_type
from ...integration.integration_metadata.constants import EXPIRE_GRACE_PERIOD_DEFAULT
from ...integration.integration_metadata.fields import (
    validate_emojis_enabled, validate_expire_behavior, validate_expire_grace_period
)

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..conversion_helpers.converters import value_deserializer_name, value_serializer_name


# ---- emojis_enabled ----

EMOJIS_ENABLED_CONVERSION = AuditLogEntryChangeConversion(
    ('enable_emoticons',),
    'emojis_enabled',
    value_validator = validate_emojis_enabled,
)


@EMOJIS_ENABLED_CONVERSION.set_value_deserializer
def emojis_enabled_value_deserializer(value):
    if value is None:
        value = True
    return value


# ---- expire_behavior ----

EXPIRE_BEHAVIOR_CONVERSION = AuditLogEntryChangeConversion(
    ('expire_behavior',),
    'expire_behavior',
    value_validator = validate_expire_behavior,
)


@EXPIRE_BEHAVIOR_CONVERSION.set_value_deserializer
def expire_behavior_value_deserializer(value):
    return IntegrationExpireBehavior(value)


@EXPIRE_BEHAVIOR_CONVERSION.set_value_serializer
def expire_behavior_value_serializer(value):
    return value.value


# ---- expire_grace_period ----

EXPIRE_GRACE_PERIOD_CONVERSION = AuditLogEntryChangeConversion(
    ('expire_grace_period',),
    'expire_grace_period',
    value_validator = validate_expire_grace_period,
)

@EXPIRE_GRACE_PERIOD_CONVERSION.set_value_deserializer
def expire_grace_period_value_deserializer(value):
    if value is None:
        value = EXPIRE_GRACE_PERIOD_DEFAULT
    return value


# ---- name ----

NAME_CONVERSION = AuditLogEntryChangeConversion(
    ('name',),
    'name',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_name,
)


# ---- type ----

TYPE_CONVERSION = AuditLogEntryChangeConversion(
    ('type',),
    'type',
    value_validator = validate_type,
)


@TYPE_CONVERSION.set_value_deserializer
def type_value_deserializer(value):
    return IntegrationType(value)


@TYPE_CONVERSION.set_value_serializer
def type_value_serializer(value):
    return value.value


# ---- Construct ----

INTEGRATION_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    EMOJIS_ENABLED_CONVERSION,
    EXPIRE_BEHAVIOR_CONVERSION,
    EXPIRE_GRACE_PERIOD_CONVERSION,
    NAME_CONVERSION,
    TYPE_CONVERSION,
)



__all__ = ()


def _nullable_list_intersection(list_0, list_1):
    """
    Returns the intersection of 2 nullable lists.
    
    Parameters
    ----------
    list_0 : `None | list`
        Input list.
    list_1 : `None | list`
        Input list.
    
    Returns
    -------
    intersection : `None | list`
        A list with the two list's intersection.
    """
    if list_0 is None:
        return None
    
    if list_1 is None:
        return None
    
    intersection = {*list_0} & {*list_1}
    if not intersection:
        return None
    
    return [*intersection]


def _nullable_list_symmetric_difference(list_0, list_1):
    """
    Returns the symmetric difference of 2 nullable lists.
    
    Parameters
    ----------
    list_0 : `None | list`
        Input list.
    list_1 : `None | list`
        Input list.
    
    Returns
    -------
    symmetric_difference : `None | list`
        A list with the two list's symmetric difference.
    """
    if list_0 is None:
        if list_1 is None:
            return None
        
        else:
            return list_1.copy()
    
    else:
        if list_1 is None:
            return list_0.copy()
    
    symmetric_difference = {*list_0} ^ {*list_1}
    if not symmetric_difference:
        return None
    
    return [*symmetric_difference]


def _nullable_list_union(list_0, list_1):
    """
    Returns the union of 2 nullable lists.
    
    Parameters
    ----------
    list_0 : `None | list`
        Input list.
    list_1 : `None | list`
        Input list.
    
    Returns
    -------
    union : `None | list`
        A list with the two list's union.
    """
    if list_0 is None:
        if list_1 is None:
            return None
        
        else:
            return list_1.copy()
    
    else:
        if list_1 is None:
            return list_0.copy()
        
        else:
            return [*{*list_0, *list_1}]


def _nullable_list_difference(list_0, list_1):
    """
    Returns the a copy of `list_0` without the elements of `list_1`.
    
    Parameters
    ----------
    list_0 : `None | list`
        Input list.
    list_1 : `None | list`
        Input list.
    
    Returns
    -------
    difference : `None | list`
        A list with the two list's difference.
    """
    if list_0 is None:
        return None
    
    if list_1 is None:
        return list_0.copy()
    
    difference = {*list_0} - {*list_1}
    if not difference:
        return None
    
    return [*difference]




__all__ = ('PollVoteAddEvent',)

from scarletio import copy_docs

from ...bases import EventBase
from ...user import create_partial_user_from_id

from .fields import (
    parse_answer_id, parse_message, parse_user_id, put_answer_id, put_message, put_user_id,
    validate_answer_id, validate_message, validate_user_id
)


class PollVoteAddEvent(EventBase):
    """
    Represents a processed `MESSAGE_POLL_VOTE_ADD` dispatch event.
    
    Attributes
    ----------
    answer_id : `int`
        The voted answer's identifier.
    message : ``Message``
        The message on what the vote is added.
    user_id : `int`
        The user's identifier who voted
    """
    __slots__ = ('answer_id', 'message', 'user_id')
    
    def __new__(cls, message, answer_id, user_id):
        """
        Creates a new vote add (or delete) instance.
        
        Parameters
        ----------
        message : ``Message``
            The respective message.
        answer_id : `int`
            The answer_id used.
        user_id : `int`
            The user_id who reacted.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # answer_id
        answer_id = validate_answer_id(answer_id)
        
        # message
        message = validate_message(message)
        
        # user_id
        user_id = validate_user_id(user_id)
        
        # Construct
        self = object.__new__(cls)
        self.answer_id = answer_id
        self.message = message
        self.user_id = user_id
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new vote add (or delete) event instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Vote event data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.answer_id = parse_answer_id(data)
        self.message = message = parse_message(data)
        self.user_id = parse_user_id(data)
        return self
    
    
    @classmethod
    def from_fields(cls, message, answer_id, user_id):
        """
        Creates a new vote add (or delete) instance.
        
        Parameters
        ----------
        message : ``Message``
            The respective message.
        answer_id : `int`
            The answer_id used.
        user_id : `int`
            The user_id who reacted.
        
        Returns
        -------
        new : `instance<cls>`
        """
        self = object.__new__(cls)
        self.answer_id = answer_id
        self.message = message
        self.user_id = user_id
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the vote add (or delete) event into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_answer_id(self.answer_id, data, defaults)
        put_message(self.message, data, defaults)
        put_user_id(self.user_id, data, defaults)
        return data
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' message = ')
        repr_parts.append(repr(self.message))
        
        repr_parts.append(', answer_id = ')
        repr_parts.append(repr(self.answer_id))
        
        repr_parts.append(', user_id = ')
        repr_parts.append(repr(self.user_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 3
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.message
        yield self.answer_id
        yield self.user_id
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.answer_id != other.answer_id:
            return False
        
        if self.message is not other.message:
            return False
        
        if self.user_id != other.user_id:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # answer_id
        hash_value ^= self.answer_id
        
        # message
        hash_value ^= hash(self.message)
        
        # user_id
        hash_value ^= self.user_id
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the vote add (or remove) event.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.answer_id = self.answer_id
        new.message = self.message
        new.user_id = self.user_id
        return new
    
    
    def copy_with(self, *, answer_id = ..., message = ..., user_id = ...):
        """
        Copies the vote add (or remove) event with the given fields.
        
        Parameters
        ----------
        answer_id : `int`, Optional (Keyword only)
            The answer_id used.
        message : ``Message``, Optional (Keyword only)
            The respective message.
        user_id : `int`, Optional (Keyword only)
            The user_id who reacted.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # answer_id
        if answer_id is ...:
            answer_id = self.answer_id
        else:
            answer_id = validate_answer_id(answer_id)
        
        # message
        if message is ...:
            message = self.message
        else:
            message = validate_message(message)
        
        # user_id
        if user_id is ...:
            user_id = self.user_id
        else:
            user_id = validate_user_id(user_id)
        
        new = object.__new__(type(self))
        new.answer_id = answer_id
        new.message = message
        new.user_id = user_id
        return new
    
    
    @property
    def user(self):
        """
        Returns the user who voted (removed their vote).
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        return create_partial_user_from_id(self.user_id)



from .application_command import *
from .application_command_option import *
from .application_command_option_choice import *
from .application_command_option_metadata import *
from .application_command_permission import *
from .application_command_permission_overwrite import *
from .helpers import *


__all__ = (
    *application_command.__all__,
    *application_command_option.__all__,
    *application_command_option_choice.__all__,
    *application_command_option_metadata.__all__,
    *application_command_permission.__all__,
    *application_command_permission_overwrite.__all__,
    *helpers.__all__,
)



__all__ = ('ScheduleNthWeeksDay',)

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_nth_week, parse_weeks_day, put_nth_week, put_weeks_day, validate_nth_week, validate_weeks_day
)
from .preinstanced import ScheduleWeeksDay


class ScheduleNthWeeksDay(RichAttributeErrorBaseType):
    """
    Represents a day of the nth week of a month.
    
    Attributes
    ----------
    nth_week : `int`
        The week of the month.
    weeks_day : ``ScheduleWeeksDay``
        The day of the week.
    """
    __slots__ = ('nth_week', 'weeks_day')
    
    def __new__(cls, *, nth_week = ..., weeks_day = ...):
        """
        Creates a new nth week's day.
        
        Parameters
        ----------
        nth_week : `int`, Optional (Keyword only)
            The week of the month.
        weeks_day : ``ScheduleWeeksDay``, `int`, Optional (Keyword only)
            The day of the week.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # nth_week
        if nth_week is ...:
            nth_week = 1
        else:
            nth_week = validate_nth_week(nth_week)
        
        # weeks_day
        if weeks_day is ...:
            weeks_day = ScheduleWeeksDay.monday
        else:
            weeks_day = validate_weeks_day(weeks_day)
        
        # construct
        self = object.__new__(cls)
        self.nth_week = nth_week
        self.weeks_day = weeks_day
        return self
    
    
    def __repr__(self):
        """Returns the nth week's day's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # nth_week
        repr_parts.append(' nth_week = ')
        repr_parts.append(repr(self.nth_week))
        
        # weeks_day
        repr_parts.append(' weeks_day = ')
        repr_parts.append(self.weeks_day.name)
        # Skip adding " ~ value"
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two nth week's days are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # nth_week
        if self.nth_week != other.nth_week:
            return False
        
        # weeks_day
        if self.weeks_day is not other.weeks_day:
            return False
        
        return True
    
    
    def __gt__(self, other):
        """Returns whether the two nth week's days is greater than the other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # nth_week
        self_nth_week = self.nth_week
        other_nth_week = other.nth_week
        if self_nth_week > other_nth_week:
            return True
        
        if self_nth_week != other_nth_week:
            return False
        
        # weeks_day
        if self.weeks_day > other.weeks_day:
            return True
        
        return False
    
    
    def __hash__(self):
        """Returns whether the two nth week's days hash value."""
        hash_value = 0
        
        # nth_week
        hash_value ^= self.nth_week << 4
        
        # weeks_day
        hash_value ^= hash(self.weeks_day)
        
        return hash_value
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new nth week's day from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.nth_week = parse_nth_week(data)
        self.weeks_day = parse_weeks_day(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the schedule nth week's day to a serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_nth_week(self.nth_week, data, defaults)
        put_weeks_day(self.weeks_day, data, defaults)
        return data
    
    
    def copy(self):
        """
        Returns a copy of the nth week's day.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.nth_week = self.nth_week
        new.weeks_day = self.weeks_day
        return new
    
    
    def copy_with(self, *, nth_week = ..., weeks_day = ...):
        """
        Returns the copy of the nth week's day with the given fields.
        
        Parameters
        ----------
        nth_week : `int`, Optional (Keyword only)
            The week of the month.
        weeks_day : ``ScheduleWeeksDay``, `int`, Optional (Keyword only)
            The day of the week.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # nth_week
        if nth_week is ...:
            nth_week = self.nth_week
        else:
            nth_week = validate_nth_week(nth_week)
        
        # weeks_day
        if weeks_day is ...:
            weeks_day = self.weeks_day
        else:
            weeks_day = validate_weeks_day(weeks_day)
        
        # construct
        new = object.__new__(type(self))
        new.nth_week = nth_week
        new.weeks_day = weeks_day
        return new



from .application_command_option_choice import *
from .constants import *
from .fields import *


__all__ = (
    *application_command_option_choice.__all__,
    *constants.__all__,
    *fields.__all__,
)



__all__ = ('ComponentMetadataChannelSelect', )

from scarletio import copy_docs

from .entity_select_base import ComponentMetadataEntitySelectBase
from .fields import parse_channel_types, put_channel_types, validate_channel_types


class ComponentMetadataChannelSelect(ComponentMetadataEntitySelectBase):
    """
    Channel select component metadata.
    
    Attributes
    ----------
    channel_types : `None`, `tuple` of ``ChannelType``
        The allowed channel types by the select.
    
    custom_id : `None`, `str`
        Custom identifier to detect which component was used by the user.
    
    default_values : `None`, `tuple` of ``EntitySelectDefaultValue``
        Entities presented in the select by default.
    
    enabled : `bool`
        Whether the component is enabled.
    
    max_values : `int
        The maximal amount of options to select.
    
    min_values : `int`
        The minimal amount of options to select.
    
    placeholder : `None`, `str`
        Placeholder text of the select.
    """
    __slots__ = ('channel_types',)
    
    
    def __new__(
        cls,
        *,
        channel_types = ...,
        custom_id = ...,
        default_values = ...,
        enabled = ...,
        max_values = ...,
        min_values = ...,
        placeholder = ...,
    ):
        """
        Creates a new channel select component metadata with the given parameters.
        
        Parameters
        ----------
        channel_types : `None`, `iterable` of (``ChannelType``, `int`), Optional (Keyword only)
            The allowed channel types by the select.
        
        custom_id : `None`, `str`, Optional (Keyword only)
            Custom identifier to detect which component was used by the user.
        
        default_values : `None | iterable<Channel | Role | ClientUserBase | EntitySelectDefaultValue | tuple>` \
                , Optional (Keyword only)
            Entities presented in the select by default.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the component is enabled.
        
        max_values : `int, Optional (Keyword only)
            The maximal amount of options to select.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of options to select.
        
        placeholder : `None`, `str`, Optional (Keyword only)
            Placeholder text of the select.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_types
        if channel_types is ...:
            channel_types = None
        else:
            channel_types = validate_channel_types(channel_types)
        
        # Construct
        self = ComponentMetadataEntitySelectBase.__new__(
            cls,
            custom_id = custom_id,
            default_values = default_values,
            enabled = enabled,
            max_values = max_values,
            min_values = min_values,
            placeholder = placeholder,
        )
        self.channel_types = channel_types
        
        return self
    
    
    @classmethod
    @copy_docs(ComponentMetadataEntitySelectBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            channel_types = keyword_parameters.pop('channel_types', ...),
            custom_id = keyword_parameters.pop('custom_id', ...),
            default_values = keyword_parameters.pop('default_values', ...),
            enabled = keyword_parameters.pop('enabled', ...),
            max_values = keyword_parameters.pop('max_values', ...),
            min_values = keyword_parameters.pop('min_values', ...),
            placeholder = keyword_parameters.pop('placeholder', ...),
        )
    
    
    @copy_docs(ComponentMetadataEntitySelectBase._add_type_specific_repr_fields_into)
    def _add_type_specific_repr_fields_into(self, repr_parts):
        # channel_types
        repr_parts.append(', channel_types = ')
        channel_types = self.channel_types
        if (channel_types is None):
            repr_parts.append('[]')
        else:
            repr_parts.append('[')
            
            index = 0
            limit = len(channel_types)
            
            while True:
                option = channel_types[index]
                index += 1
                
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
    
    
    @copy_docs(ComponentMetadataEntitySelectBase.__hash__)
    def __hash__(self):
        hash_value = ComponentMetadataEntitySelectBase.__hash__(self)
        
        # channel_types
        channel_types = self.channel_types
        if (channel_types is not None):
            hash_value ^= len(channel_types) << 12
            for index, channel_type in enumerate(channel_types):
                hash_value ^= channel_type.value << (index << 1)
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataEntitySelectBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ComponentMetadataEntitySelectBase._is_equal_same_type(self, other):
            return False
        
        # channel_types
        if self.channel_types != other.channel_types:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataEntitySelectBase.from_data)
    def from_data(cls, data):
        self = super(ComponentMetadataChannelSelect, cls).from_data(data)
        self.channel_types = parse_channel_types(data)
        return self
    
    
    @copy_docs(ComponentMetadataEntitySelectBase.to_data)
    def to_data(self, *, defaults = False):
        data = ComponentMetadataEntitySelectBase.to_data(self)
        
        put_channel_types(self.channel_types, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataEntitySelectBase.copy)
    def copy(self):
        new = ComponentMetadataEntitySelectBase.copy(self)
        
        # channel_types
        channel_types = self.channel_types
        if (channel_types is not None):
            channel_types = (*channel_types,)
        new.channel_types = channel_types
        
        return new
    
    
    def copy_with(
        self,
        *,
        channel_types = ...,
        custom_id = ...,
        default_values = ...,
        enabled = ...,
        max_values = ...,
        min_values = ...,
        placeholder = ...,
    ):
        """
        Copies the channel select component metadata with the given fields.
        
        Parameters
        ----------
        channel_types : `None`, `iterable` of (``ChannelType``, `int`), Optional (Keyword only)
            The allowed channel types by the select.
        
        custom_id : `None`, `str`, Optional (Keyword only)
            Custom identifier to detect which component was used by the user.
        
        default_values : `None | iterable<Channel | Role | ClientUserBase | EntitySelectDefaultValue | tuple>` \
                , Optional (Keyword only)
            Entities presented in the select by default.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the component is enabled.
        
        max_values : `int, Optional (Keyword only)
            The maximal amount of options to select.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of options to select.
        
        placeholder : `None`, `str`, Optional (Keyword only)
            Placeholder text of the select.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_types
        if channel_types is ...:
            channel_types = self.channel_types
            if (channel_types is not None):
                channel_types = (*channel_types,)
        else:
            channel_types = validate_channel_types(channel_types)
        
        # Construct
        
        new = ComponentMetadataEntitySelectBase.copy_with(
            self,
            custom_id = custom_id,
            default_values = default_values,
            enabled = enabled,
            max_values = max_values,
            min_values = min_values,
            placeholder = placeholder,
        )
        new.channel_types = channel_types
        return new
    
    
    @copy_docs(ComponentMetadataEntitySelectBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            channel_types = keyword_parameters.pop('channel_types', ...),
            custom_id = keyword_parameters.pop('custom_id', ...),
            default_values = keyword_parameters.pop('default_values', ...),
            enabled = keyword_parameters.pop('enabled', ...),
            max_values = keyword_parameters.pop('max_values', ...),
            min_values = keyword_parameters.pop('min_values', ...),
            placeholder = keyword_parameters.pop('placeholder', ...),
        )



__all__ = ('CONVERSION_FORWARD_MESSAGE',)

from ....bases import maybe_snowflake_pair
from ....builder.constants import CONVERSION_KIND_FIELD
from ....builder.conversion import Conversion

from ...message import Message

from ..message_reference_configuration import MessageReferenceConfiguration
from ..preinstanced import MessageReferenceType

from .message_reference_configuration import CONVERSION_MESSAGE_REFERENCE_CONFIGURATION


class CONVERSION_FORWARD_MESSAGE(Conversion):
    # Generic
    
    name = 'forward_message'
    name_aliases = None
    expected_types_messages = f'`None`, `{Message.__name__}`, `(int, int)`'
    kind = CONVERSION_KIND_FIELD
    output_conversion = CONVERSION_MESSAGE_REFERENCE_CONFIGURATION
    
    # Setting
    
    set_merger = None
    set_type = None
    set_type_processor = None
    set_identifier = None
    set_listing_identifier = None
    
    
    def set_validator(value):
        # None
        if value is None:
            yield MessageReferenceConfiguration()
            return
        
        # Message
        if isinstance(value, Message):
            yield MessageReferenceConfiguration(
                channel_id = value.channel_id,
                message_id = value.id,
                message_reference_type = MessageReferenceType.forward,
            )
            return
        
        # (int, int)
        snowflake_pair = maybe_snowflake_pair(value)
        if (snowflake_pair is not None):
            yield MessageReferenceConfiguration(
                channel_id = snowflake_pair[0],
                message_id = snowflake_pair[1],
                message_reference_type = MessageReferenceType.forward,
            )
            return
        
        # No other cases
        return
    
    # Reading
    
    get_default = None
    
    def get_processor(value):
        if value is None:
            return None
        
        return (value.channel_id, value.message_id)
    
    # Serialization
    
    serializer_key = None
    serializer_optional = None
    serializer_required = None
    
    # Sorting
    
    sort_priority = 2300



__all__ = ()

from scarletio import KeepType

from ...discord.client.client_wrapper import ClientWrapper
from ...discord.events.handling_helpers import _EventHandlerManagerRouter, Router

from .command_processor import CommandProcessor


def commands_getter(manager_router):
    """
    Gets the command processor using `Client.command_processor` of an ``_EventHandlerManagerRouter``.
    
    Parameters
    ----------
    manager_router : ``_EventHandlerManagerRouter``
        The caller manager router.
    
    Returns
    -------
    handlers : `list` of ``CommandProcessor``
    """
    handlers = []
    for client in manager_router.parent.clients:
        manager = getattr(client, 'commands', None)
        if manager is None:
            continue
        
        handler = manager.parent
        if isinstance(handler, CommandProcessor):
            handlers.append(handler)
    
    return handlers


@KeepType(ClientWrapper)
class ClientWrapper:
    
    @property
    def commands(self):
        """
        Returns a ``_EventHandlerManagerRouter``, with what commands can be added to more clients at the
        same time.
        
        Returns
        -------
        event_handler_manager_router : ``_EventHandlerManagerRouter``
        """
        return _EventHandlerManagerRouter(self, commands_getter, Router)



__all__ = ()

from ...emoji import Emoji, create_emoji_from_exclusive_inline_data, put_exclusive_emoji_inline_data_into
from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, float_parser_factory, force_string_parser_factory,
    nullable_entity_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_optional_putter_factory, entity_id_putter_factory,
    float_optional_putter_factory, force_string_putter_factory, nullable_entity_optional_putter_factory,
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, float_conditional_validator_factory,
    force_string_validator_factory, nullable_entity_validator_factory
)
from ...user import ClientUserBase, User

from .constants import NAME_LENGTH_MAX, NAME_LENGTH_MIN


# Unused fields:
# id -> use `sound_id`. Redundant with `sound_id`.
# override_path -> only present in builtin sounds. Redundant with `name`.

# available

parse_available = bool_parser_factory('available', True)
put_available = bool_optional_putter_factory('available', True)
validate_available = bool_validator_factory('available', True)

# emoji

def parse_emoji(data):
    """
    Parses out the `emoji` field from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Channel data.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    return create_emoji_from_exclusive_inline_data(data)



def put_emoji(emoji, data, defaults):
    """
    Puts the `emoji`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    emoji : `None`, ``Emoji``
        The forum tag's emoji.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if defaults or (emoji is not None):
        put_exclusive_emoji_inline_data_into(emoji, data)
    
    return data


validate_emoji = nullable_entity_validator_factory('emoji', Emoji)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# id

parse_id = entity_id_parser_factory('sound_id')
put_id = entity_id_putter_factory('sound_id')
validate_id = entity_id_validator_factory('sound_id')

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# user

parse_user = nullable_entity_parser_factory('user', User)
put_user = nullable_entity_optional_putter_factory('user', ClientUserBase, force_include_internals = True)
validate_user = nullable_entity_validator_factory('user', ClientUserBase)

# user_id

parse_user_id = entity_id_parser_factory('user_id')
put_user_id = entity_id_putter_factory('user_id')
validate_user_id = entity_id_validator_factory('user_id', ClientUserBase)

# volume

parse_volume = float_parser_factory('volume', 1.0)
put_volume = float_optional_putter_factory('volume', 1.0)
validate_volume = float_conditional_validator_factory(
    'volume',
    1.0,
    lambda volume : volume >= 0.0 and volume <= 1.0,
    '>= 0.0 and <= 1.0',
)



from .embed import *
from .embed_author import *
from .embed_field import *
from .embed_field_base import *
from .embed_footer import *
from .embed_image import *
from .embed_provider import *
from .embed_thumbnail import *
from .embed_video import *


__all__ = (
    *embed.__all__,
    *embed_author.__all__,
    *embed_field.__all__,
    *embed_field_base.__all__,
    *embed_footer.__all__,
    *embed_image.__all__,
    *embed_provider.__all__,
    *embed_thumbnail.__all__,
    *embed_video.__all__,
)


# Deprecations

from ...utils.module_deprecation import deprecated_import

deprecated_import(Embed, 'EmbedBase')
deprecated_import(Embed, 'EmbedCore')

del deprecated_import



__all__ = ()

FLAG_HAS_BEFORE = 1 << 0
FLAG_HAS_AFTER = 1 << 1


FLAG_NAMES = (
    (FLAG_HAS_BEFORE, 'has_before'),
    (FLAG_HAS_AFTER, 'has_after'),
)


def get_flags_name(flags):
    """
    Gets the name of the given flags.
    
    Parameters
    ----------
    flags : `int`
        Flags to get their name of.
    
    Returns
    -------
    flags_name : `str`
    """
    if not flags:
        return 'none'
    
    return ', '.join([flag_name for flag_mask, flag_name in FLAG_NAMES if flags & flag_mask])



__all__ = ('register_plugin',)

from scarletio import copy_docs

from ..plugin_loader import PLUGIN_LOADER


@copy_docs(PLUGIN_LOADER.register)
def register_plugin(name, *positional_parameters, **keyword_parameters):
    return PLUGIN_LOADER.register(name, *positional_parameters, **keyword_parameters)



__all__ = ()

def _merge_list_groups(group_1, group_2):
    """
    Mergers the two list groups.
    
    Merging groups is a pre-operation of difference calculation, so copying is not required.
    
    Parameters
    ----------
    group_1 : `None`, `list` of `object`
        The first group to merge.
    group_2 : `None`, `list` of `object`
        The second group to merge.
    
    Returns
    -------
    merged_group : `None`, `list` of `object`
        The merged groups.
    """
    if group_1 is None:
        return group_2
    
    if group_2 is None:
        return group_1
    
    merged_group = [*group_1]
    for value in group_2:
        if (value not in merged_group):
            merged_group.append(value)
    
    return merged_group


def _merge_set_groups(group_1, group_2):
    """
    Mergers the two set groups.
    
    Merging groups is a pre-operation of difference calculation, so copying is not required.
    
    Parameters
    ----------
    group_1 : `None`, `set` of `object`
        The first group to merge.
    group_2 : `None`, `set` of `object`
        The second group to merge.
    
    Returns
    -------
    merged_group : `None`, `set` of `object`
        The merged groups.
    """
    if group_1 is None:
        return group_2
    
    if group_2 is None:
        return group_1
    
    return {*group_1, *group_2}


def _get_list_difference(group_1, group_2):
    """
    Calculates the difference between the two event value group.
    
    The returned lists are always a copy of the original.
    
    Parameters
    ----------
    group_1 : `None`, `list` of `object`
        The first group.
    group_2 : `None`, `list` of `object`
        The second group.
    
    Returns
    -------
    group_1 : `None`, `list` of `object`
        The first group.
    group_2 : `None`, `list` of `object`
        The second group.
    """
    not_null = 0
    
    if (group_1 is not None):
        group_1 = group_1.copy()
        not_null += 1
        
    if (group_2 is not None):
        group_2 = group_2.copy()
        not_null += 1
    
    if not_null == 2:
        for index in reversed(range(len(group_1))):
            value = group_1[index]
            try:
                group_2.remove(value)
            except ValueError:
                pass
            else:
                del group_1[index]
        
        if not group_2:
            group_2 = None
        
        if not group_1:
            group_1 = None
    
    return group_1, group_2


def _get_set_difference(group_1, group_2):
    """
    Calculates the difference between the two event value group.
    
    The returned sets are always a copy of the original.
    
    Parameters
    ----------
    group_1 : `None`, `set` of `object`
        The first group.
    group_2 : `None`, `set` of `object`
        The second group.
    
    Returns
    -------
    new_group_1 : `None`, `set` of `object`
        The first group.
    new_group_2 : `None`, `set` of `object`
        The second group.
    """
    if (group_1 is None) or (group_2 is None):
        if (group_1 is None):
            new_group_1 = None
        else:
            new_group_1 = group_1.copy()
            
        if (group_2 is None):
            new_group_2 = None
        else:
            new_group_2 = group_2.copy()
    
    else:
        new_group_1 = group_1 - group_2
        new_group_2 = group_2 - group_1
        
        if not new_group_1:
            new_group_1 = None
        
        if not new_group_2:
            new_group_2 = None
    
    return new_group_1, new_group_2



__all__ = ()

from scarletio import include_with_callback, set_docs

from .utils import timestamp_to_datetime


def entity_id_parser_factory(field_key):
    """
    Returns a new entity id parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out an entity id field from the given data.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        entity_id : `int`
        """
        nonlocal field_key
            
        entity_id = data.get(field_key, None)
        if (entity_id is None):
            entity_id = 0
        else:
            entity_id = int(entity_id)
        
        return entity_id
    
    return parser


def entity_id_array_parser_factory(field_key):
    """
    Returns a new entity id array parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out an entity id field from the given data.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        entity_id_array : `None`, `tuple` of `int`
        """
        nonlocal field_key
                
        entity_id_array = data.get(field_key, None)
        if (entity_id_array is None) or (not entity_id_array):
            entity_id_array = None
        else:
            entity_id_array = tuple(sorted(int(entity_id) for entity_id in entity_id_array))
        
        return entity_id_array
    
    return parser


def preinstanced_parser_factory(
    field_key, preinstanced_type, default_value, *, include = None, include_default_attribute_name = None
):
    """
    Returns a new preinstanced object parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    preinstanced_type : ``PreinstancedBase``
        The preinstanced type to use.
    
    default_value : `instance<preinstanced_type>`, `object`
        The default value to use if the key is not present.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The type's name to include `preinstanced_type` with.
        Should be used when `preinstanced_type` cannot be resolved initially.
    
    include_default_attribute_name : `None`, `str` = `None`, Optional (Keyword only)
        If `preinstanced_type` is included meaning `default_value` cannot be initially defined for this reason, with
        this parameter it is possible to define which attribute of `preinstanced_type` should be pulled.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a preinstanced field from the given data.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        preinstanced : ``PreinstancedBase``
        """
        nonlocal field_key
        nonlocal default_value
        nonlocal preinstanced_type
        
        try:
            value = data[field_key]
        except KeyError:
            preinstanced = default_value
        else:
            preinstanced = preinstanced_type(value)
        
        return preinstanced
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal preinstanced_type
            nonlocal default_value
            nonlocal include_default_attribute_name
            
            preinstanced_type = value
            
            if (include_default_attribute_name is not None):
                default_value = getattr(preinstanced_type, include_default_attribute_name)
    
    return parser


def preinstanced_array_parser_factory(field_key, preinstanced_type, *, include = None):
    """
    Returns a new preinstanced array parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
        
    preinstanced_type : ``PreinstancedBase``
        The preinstanced type to use.

    include : `None`, `str` = `None`, Optional (Keyword only)
        The type's name to include `preinstanced_type` with.
        Should be used when `preinstanced_type` cannot be resolved initially.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a preinstanced field from the given data.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        preinstanced_array : `None`, `tuple` of ``PreinstancedBase``
        """
        nonlocal field_key
        nonlocal preinstanced_type
        
        value_array = data.get(field_key, None)
        if (value_array is None) or (not value_array):
            preinstanced_array = None
        else:
            preinstanced_array = tuple(sorted(preinstanced_type(value) for value in value_array))
        
        return preinstanced_array
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal preinstanced_type
            preinstanced_type = value
    
    return parser


def bool_parser_factory(field_key, default_value):
    """
    Returns a `bool` parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `bool`
        The default value to use if the key is not present.
    
    Returns
    -------
    parser : `FunctionType`
    """
    return _field_parser_factory(field_key, default_value)


def float_parser_factory(field_key, default_value):
    """
    Returns an `float` parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `float`
        The default value to use if the key is not present.
    
    Returns
    -------
    parser : `FunctionType`
    """
    return _field_parser_factory(field_key, default_value)


def int_parser_factory(field_key, default_value):
    """
    Returns an `int` parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `int`
        The default value to use if the key is not present.
    
    Returns
    -------
    parser : `FunctionType`
    """
    return _field_parser_factory(field_key, default_value)


def int_postprocess_parser_factory(field_key, default_value, postprocessor):
    """
    Returns an `int` parser with a postprocessor function applied to it.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `int`
        The default value to use if the key is not present.
    postprocessor : `callable`
        Postprocessor to call on the field.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out an integer from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `int`
        """
        nonlocal field_key
        nonlocal default_value
        nonlocal postprocessor
        
        value = data.get(field_key, None)
        if (value is None):
            value = default_value
        else:
            value = postprocessor(value)
        
        return value
    
    return parser


def flag_parser_factory(field_key, flag_type, *, default_value = ...):
    """
    Returns a `flag` parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    flag_type : `type`
        The type of the flag to return.
    default_value : `instance<flag_type>`, Optional (Keyword only)
        Default value to return if `None` is received.
    
    Returns
    -------
    parser : `FunctionType`
    """
    if default_value is ...:
        default_value = flag_type()
    
    def parser(data):
        """
        Parses out an flag from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `instance<flag_type>`
        """
        nonlocal default_value
        nonlocal field_key
        nonlocal flag_type
        
        value = data.get(field_key, None)
        if (value is None):
            value = default_value
        else:
            value = flag_type(value)
        
        return value
    
    return parser


def nullable_flag_parser_factory(field_key, flag_type):
    """
    Returns a nullable `flag` parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    flag_type : `type`
        The type of the flag to return.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a nullable flag from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `None`, `instance<flag_type>`
        """
        nonlocal field_key
        nonlocal flag_type
        
        value = data.get(field_key, None)
        if (value is not None):
            value = flag_type(value)
        
        return value
    
    return parser



def negated_bool_parser_factory(field_key, default_value):
    """
    Returns a negated `bool` parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `bool`
        The default value to use if the key is not present.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a `bool` from the given payload as returns it as negated.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `bool`
        """
        nonlocal field_key
        nonlocal default_value
        
        value = data.get(field_key, None)
        if value is None:
            value = default_value
        else:
            value = not value
        
        return value
    
    return parser


def nullable_date_time_parser_factory(field_key):
    """
    Returns a new nullable date time parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    parser : `FunctionType`
    """
    return default_date_time_parser_factory(field_key, None)


def default_date_time_parser_factory(field_key, default):
    """
    Returns a new defaulted date time parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default : `object`
        Default value to return if the field is not present (or null) in the data.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a date time from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `None`, `DateTime`
        """
        nonlocal default
        nonlocal field_key
        
        timestamp = data.get(field_key, None)
        if (timestamp is not None):
            return timestamp_to_datetime(timestamp)
        
        return default
    
    return parser


def force_string_parser_factory(field_key):
    """
    Returns a new forced string parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a string from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `str`
        """
        nonlocal field_key
        
        field_value = data.get(field_key, None)
        if (field_value is None):
            field_value = ''
        
        return field_value
    
    return parser


def field_parser_factory(field_key):
    """
    Returns a field parser. This parser 1:1 returns the parsed value.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out any value from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `object`
        """
        nonlocal field_key
        
        field_value = data.get(field_key, None)
        if (field_value is not None) and isinstance(field_value, str) and (not field_value):
            field_value = None
        
        return field_value
    
    return parser


def _field_parser_factory(field_key, default_value):
    """
    Returns an `object` parser. If the parsed object is `None` returns the given `default_value`.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    default_value : `object`
        The default value to use if the key is not present.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a nullable field from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `int`
        """
        nonlocal field_key
        nonlocal default_value
        
        value = data.get(field_key, None)
        if (value is None):
            value = default_value
        
        return value
    
    return parser



def nullable_string_parser_factory(field_key):
    """
    Returns a nullable string parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a string from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `str`
        """
        nonlocal field_key
        
        field_value = data.get(field_key, None)
        if (field_value is not None) and (not field_value):
            field_value = None
        
        return field_value
    
    return parser


def nullable_array_parser_factory(field_key, *, ordered = True):
    """
    Returns a nullable sorted object array parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    ordered : `bool` = `True`, Optional (Keyword only)
        Whether the array should be sorted.
    
    Returns
    -------
    parser : `FunctionType`
    """
    if ordered:
        def parser(data):
            """
            Parses out a nullable sorted object array from the given payload.
            
            > This function is generated.
            
            Parameters
            ----------
            data : `dict<str, object>`
                Data to parse from.
            
            Returns
            -------
            sorted_array : `None | tuple<object>`
            """
            nonlocal field_key
            
            raw_object_array = data.get(field_key, None)
            if (raw_object_array is None) or (not raw_object_array):
                object_array = None
            else:
                object_array = tuple(sorted(raw_object_array))
            
            return object_array
    
    else:
        def parser(data):
            """
            Parses out a nullable object array from the given payload.
            
            > This function is generated.
            
            Parameters
            ----------
            data : `dict<str, object>`
                Data to parse from.
            
            Returns
            -------
            sorted_array : `None | tuple<object>`
            """
            nonlocal field_key
            
            raw_object_array = data.get(field_key, None)
            if (raw_object_array is None) or (not raw_object_array):
                object_array = None
            else:
                object_array = tuple(raw_object_array)
            
            return object_array
    
    return parser


def nullable_sorted_array_postprocess_parser_factory(field_key, postprocessor):
    """
    Returns a nullable sorted object array parser with postprocessor applied on each value.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    postprocessor : `callable`
        Postprocessor to call on the field.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a nullable sorted object array from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        sorted_array : `None`, `tuple` of `object`
        """
        nonlocal field_key
        nonlocal postprocessor
        
        raw_object_array = data.get(field_key, None)
        if (raw_object_array is None) or (not raw_object_array):
            object_array = None
        else:
            object_array = tuple(sorted(postprocessor(value) for value in raw_object_array))
        
        return object_array
    
    return parser


def nullable_entity_array_parser_factory(field_key, entity_type, *, include = None):
    """
    Returns a new nullable entity array parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    entity_type : `type` with `{from_data}`
        Entity's type.
    include : `None`, `str` = `None`, Optional (Keyword only)
        The name to include `entity_type` with. Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a nullable entity array from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        entity_array : `None`, `tuple` of `instance<entity_type>`
        """
        nonlocal field_key
        nonlocal entity_type
        
        entity_data_array = data.get(field_key, None)
        if (entity_data_array is None) or (not entity_data_array):
            entity_array = None
        else:
            entity_array = tuple(sorted(entity_type.from_data(entity_data) for entity_data in entity_data_array))
        
        return entity_array
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal entity_type
            entity_type = value
    
    return parser


def nullable_entity_parser_factory(field_key, entity_type, *, include = None):
    """
    Returns an nullable entity parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    entity_type : `type` with `{from_data}`
        Entity's type.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The `entity_type`'s name to include.. Should be used when `entity_type` cannot be resolved initially.
    
    Returns
    -------
    parser : `FunctionType`
    """
    return default_entity_parser_factory(field_key, entity_type, default = None, include = include)


def default_entity_parser_factory(field_key, entity_type, *, default = ..., default_factory = ..., include = None):
    """
    Returns an entity parser with default return value.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    entity_type : `type` with `{from_data}`
        Entity's type.
    
    default : `object`, Optional (Keyword only)
        Default value to return if the entity is missing from the payload.
        
        Mutually exclusive with `default_factory` and at least one of them is required.
    
    default_factory : `() -> object`, Optional (Keyword only)
        Default value factory to call when the entity is missing from the payload.
        
        Mutually exclusive with `default` and at least one of them is required.
    
    Returns
    -------
    parser : `(dict<str, object>) -> instance<entity_type> | default | return<default_factory>`
    """
    if default is ... and default_factory is ...:
        raise TypeError(
            f'Either `default` or `default_factory` are required.'
        )
    
    if default is not ... and default_factory is not ...:
        raise TypeError(
            f'`default` and `default_factory` are mutually exclusive. '
            f'Got default = {default!r}, default_factory = {default_factory!r}.'
        )
    
    
    if default is not ...:
        def parser(data):
            nonlocal default
            nonlocal field_key
            nonlocal entity_type
            
            entity_data = data.get(field_key, None)
            if entity_data is None:
                entity = default
            else:
                entity = entity_type.from_data(entity_data)
            
            return entity
    
    else:
        def parser(data):
            nonlocal default_factory
            nonlocal field_key
            nonlocal entity_type
            
            entity_data = data.get(field_key, None)
            if entity_data is None:
                entity = default_factory()
            else:
                entity = entity_type.from_data(entity_data)
            
            return entity
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal entity_type
            entity_type = value
    
    
    set_docs(
        parser,
        """
        Parses out an entity from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        entity_array : `None`, `instance<entity_type>`
        """
    )
    
    return parser


def functional_parser_factory(field_key, function, *, include = None):
    """
    Returns a functional parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    function : `FunctionType`
        Function to call with the received field value.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The function's name to include `function` with. Should be used when `function` cannot be resolved initially.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a field from the given `data`. IF anything is received calls the specified function on it.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `None`, `object`
        """
        nonlocal field_key
        nonlocal function
        
        return function(data[field_key])
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal function
            function = value
    
    
    return parser


def nullable_functional_parser_factory(field_key, function, *, include = None):
    """
    Returns a functional parser with default return value.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    function : `FunctionType`
        Function to call with the received field value.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The function's name to include `function` with. Should be used when `function` cannot be resolved initially.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a field from the given `data`. If anything is received calls the specified function on it.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `None`, `object`
        """
        nonlocal field_key
        nonlocal function
        
        field_value = data.get(field_key, None)
        if (field_value is not None):
            field_value = function(field_value)
        
        return field_value
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal function
            function = value
    
    
    return parser


def nullable_functional_array_parser_factory(field_key, function, *, do_sort = False, include = None, sort_key = None):
    """
    Returns a functional array parser with default return value.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    function : `FunctionType`
        Function to call with the received field value.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The function's name to include `function` with. Should be used when `function` cannot be resolved initially.
    
    do_sort : `bool`
        Whether the output should be sorted.
    
    sort_key : `FunctionType`
        Sort key used to sort the entity.
    
    Returns
    -------
    parser : `FunctionType`
    """
    if do_sort:
        def parser(data):
            nonlocal field_key
            nonlocal function
            nonlocal sort_key
            
            field_values = data.get(field_key, None)
            if (field_values is None) or (not field_values):
                value = None
            else:
                value = tuple(sorted((function(field_value) for field_value in field_values), key = sort_key))
            
            return value
    else:
        def parser(data):
            nonlocal field_key
            nonlocal function
            
            field_values = data.get(field_key, None)
            if (field_values is None) or (not field_values):
                value = None
            else:
                value = (*(function(field_value) for field_value in field_values),)
            
            return value
            
    set_docs(
        parser,
        """
        Parses out a field from the given `data`.
        If anything is received calls the specified function on each of its elements.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `None`, `tuple` of `object`
        """
    )
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal function
            function = value
    
    
    return parser


def nullable_object_array_parser_factory(field_key, object_type, *, include = None):
    """
    Returns a new nullable object array parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    object_type : `type` with `{from_data}`
        Object's type.
    
    include : `None`, `str` = `None`, Optional (Keyword only)
        The name to include `object_type` with. Should be used when `object_type` cannot be resolved initially.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a nullable object array from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        object_array : `None`, `tuple` of `instance<object_type>`
        """
        nonlocal field_key
        nonlocal object_type
        
        object_data_array = data.get(field_key, None)
        if (object_data_array is None) or (not object_data_array):
            object_array = None
        else:
            object_array = (*(object_type.from_data(object_data) for object_data in object_data_array),)
        
        return object_array
    
    
    if (include is not None):
        @include_with_callback(include)
        def include_object_type(value):
            nonlocal object_type
            object_type = value
    
    
    return parser


def nullable_int_parser_factory(field_key):
    """
    Returns a nullable `int` parser.
    
    Parameters
    ----------
    field_key : `str`
        The field's key used in payload.
    
    Returns
    -------
    parser : `FunctionType`
    """
    def parser(data):
        """
        Parses out a nullable integer from the given payload.
        
        > This function is generated.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entity data.
        
        Returns
        -------
        field_value : `None`, `int`
        """
        nonlocal field_key
        
        return data.get(field_key, None)
    
    return parser



__all__ = ()

from ...channel import PermissionOverwriteTargetType
from ...channel.permission_overwrite.fields import (
    validate_allow, validate_deny, validate_target_id, validate_target_type
)
from ...permission import Permission
from ...permission.constants import PERMISSION_ALLOW_KEY, PERMISSION_DENY_KEY

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..conversion_helpers.converters import value_deserializer_id, value_serializer_id


# ---- allow ----

ALLOW_CONVERSION = AuditLogEntryChangeConversion(
    (PERMISSION_ALLOW_KEY,),
    'allow',
    value_validator = validate_allow,
)


# ---- deny ----

DENY_CONVERSION = AuditLogEntryChangeConversion(
    (PERMISSION_DENY_KEY,),
    'deny',
    value_validator = validate_deny,
)


# ----

@ALLOW_CONVERSION.set_value_deserializer
@DENY_CONVERSION.set_value_deserializer
def permission_value_deserializer(value):
    if value is None:
        value = Permission()
    else:
        value = Permission(value)
    return value


@ALLOW_CONVERSION.set_value_serializer
@DENY_CONVERSION.set_value_serializer
def permission_value_serializer(value):
    return format(value, 'd')


# ---- target_id ----

TARGET_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('id',),
    'target_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_target_id,
)


# ---- target_type ----

TARGET_TYPE_CONVERSION = AuditLogEntryChangeConversion(
    ('type',),
    'target_type',
    value_validator = validate_target_type,
)


@TARGET_TYPE_CONVERSION.set_value_deserializer
def target_type_value_deserializer(value):
    if (value is not None) and (PermissionOverwriteTargetType.VALUE_TYPE is int):
        value = int(value)
    
    return PermissionOverwriteTargetType(value)


@TARGET_TYPE_CONVERSION.set_value_serializer
def target_type_value_serializer(value):
    return str(value.value)


# ---- Construct ----

CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    ALLOW_CONVERSION,
    DENY_CONVERSION,
    TARGET_ID_CONVERSION,
    TARGET_TYPE_CONVERSION,
)



__all__ = ('register_and_load_plugin',)

from scarletio import copy_docs

from ..plugin_loader import PLUGIN_LOADER


@copy_docs(PLUGIN_LOADER.register_and_load)
def register_and_load_plugin(name, *parameters, blocking = True, **keyword_parameters):
    return PLUGIN_LOADER.register_and_load(name, *parameters, blocking = blocking, **keyword_parameters)



from .constants import *
from .fields import *
from .preinstanced import *
from .verification_screen_step import *


__all__ = (
    *constants.__all__,
    *fields.__all__,
    *preinstanced.__all__,
    *verification_screen_step.__all__,
)



__all__ = ('ChannelMetadataGuildVoice',)

from scarletio import copy_docs

from ...permission import Permission
from ...permission.permission import (
    PERMISSION_DENIED_FOR_GUILD_VOICE, PERMISSION_MASK_CONNECT, PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE,
    PERMISSION_VOICE_DENY_CONNECTION
)

from .fields import (
    parse_nsfw, parse_status, parse_video_quality_mode, put_nsfw, put_status, put_video_quality_mode,
    validate_nsfw, validate_status, validate_video_quality_mode
)
from .guild_voice_base import ChannelMetadataGuildVoiceBase
from .preinstanced import VideoQualityMode


class ChannelMetadataGuildVoice(ChannelMetadataGuildVoiceBase):
    """
    Guild voice channel metadata.
    
    Attributes
    ----------
    _cache_permission : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    bitrate : `int`
        The bitrate (in bits) of the voice channel.
    name : `str`
        The channel's name.
    nsfw : `bool`
        Whether the channel is marked as non safe for work.
    parent_id : `int`
        The channel's parent's identifier.
    permission_overwrites :`None`,  `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    region : ``VoiceRegion``
        The voice region of the channel.
    status : `None`, `str`
        The voice channel's status.
    user_limit : `int`
        The maximal amount of users, who can join the voice channel, or `0` if unlimited.
    video_quality_mode : ``VideoQualityMode``
        The video quality of the voice channel.
    
    Class Attributes
    ----------------
    order_group: `int` = `2`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('nsfw', 'status', 'video_quality_mode',)
    
    
    def __new__(
        cls,
        *,
        bitrate = ...,
        name = ...,
        nsfw = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
        region = ...,
        status = ...,
        user_limit = ...,
        video_quality_mode = ...,
    ):
        """
        Creates a new guild voice channel metadata from the given parameters.
        
        Parameters
        ----------
        bitrate : `int`, Optional (Keyword only)
            The bitrate (in bits) of the voice channel.
        name : `str`, Optional (Keyword only)
            The channel's name.
        nsfw : `bool`, Optional (Keyword only)
            Whether the channel is marked as non safe for work.
        parent_id : `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        permission_overwrites : `None`, `iterable` of ``PermissionOverwrite``, Optional (Keyword only)
            The channel's permission overwrites.
        position : `int`, Optional (Keyword only)
            The channel's position.
        region : ``VoiceRegion``, `str`, Optional (Keyword only)
            The voice region of the channel.
        status : `None`, `str`, Optional (Keyword only)
            The voice channel's status.
        user_limit : `int`, Optional (Keyword only)
            The maximal amount of users, who can join the voice channel, or `0` if unlimited.
        video_quality_mode : ``VideoQualityMode``, `int`, Optional (Keyword only)
            The video quality of the voice channel.
            
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # nsfw
        if nsfw is ...:
            nsfw = False
        else:
            nsfw = validate_nsfw(nsfw)
        
        # status
        if status is ...:
            status = None
        else:
            status = validate_status(status)
        
        # video_quality_mode
        if video_quality_mode is ...:
            video_quality_mode = VideoQualityMode.none
        else:
            video_quality_mode = validate_video_quality_mode(video_quality_mode)
        
        # Construct
        self = ChannelMetadataGuildVoiceBase.__new__(
            cls,
            bitrate = bitrate,
            name = name,
            parent_id = parent_id,
            permission_overwrites = permission_overwrites,
            position = position,
            region = region,
            user_limit = user_limit,
        )
        self.nsfw = nsfw
        self.status = status
        self.video_quality_mode = video_quality_mode
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildVoiceBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            bitrate = keyword_parameters.pop('bitrate', ...),
            name = keyword_parameters.pop('name', ...),
            nsfw = keyword_parameters.pop('nsfw', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
            region = keyword_parameters.pop('region', ...),
            status = keyword_parameters.pop('status', ...),
            user_limit = keyword_parameters.pop('user_limit', ...),
            video_quality_mode = keyword_parameters.pop('video_quality_mode', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataGuildVoiceBase.__hash__(self)
        
        # nsfw
        hash_value ^= self.nsfw << 28
        
        # status
        status = self.status
        if (status is not None):
            hash_value ^= hash(status)
        
        # video_quality_mode
        hash_value ^= self.video_quality_mode.value << 11
        
        return hash_value
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildVoiceBase.from_data)
    def from_data(cls, data):
        self = super(ChannelMetadataGuildVoice, cls).from_data(data)
        
        # status | Its only received with the initial channel payload.
        self.status = parse_status(data)
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildVoiceBase._is_equal_same_type(self, other):
            return False
        
        # nsfw
        if self.nsfw != other.nsfw:
            return False
        
        # status
        if self.status != other.status:
            return False
        
        # video_quality_mode
        if self.video_quality_mode is not other.video_quality_mode:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._get_display_name)
    def _get_display_name(self):
        return self.name.title()
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildVoiceBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildVoice, cls)._create_empty()
        
        self.nsfw = False
        self.status = None
        self.video_quality_mode = VideoQualityMode.none
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase.copy)
    def copy(self):
        new = ChannelMetadataGuildVoiceBase.copy(self)
        new.nsfw = self.nsfw
        new.status = self.status
        new.video_quality_mode = self.video_quality_mode
        return new
    
    
    def copy_with(
        self,
        *,
        bitrate = ...,
        name = ...,
        nsfw = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
        region = ...,
        status = ...,
        user_limit = ...,
        video_quality_mode = ...,
    ):
        """
        Copies the guild voice channel metadata from the given fields.
        
        Parameters
        ----------
        bitrate : `int`, Optional (Keyword only)
            The bitrate (in bits) of the voice channel.
        name : `str`, Optional (Keyword only)
            The channel's name.
        nsfw : `bool`, Optional (Keyword only)
            Whether the channel is marked as non safe for work.
        parent_id : `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        permission_overwrites : `None`, `iterable` of ``PermissionOverwrite``, Optional (Keyword only)
            The channel's permission overwrites.
        position : `int`, Optional (Keyword only)
            The channel's position.
        region : ``VoiceRegion``, `str`, Optional (Keyword only)
            The voice region of the channel.
        status : `None`, `str`, Optional (Keyword only)
            The voice channel's status.
        user_limit : `int`, Optional (Keyword only)
            The maximal amount of users, who can join the voice channel, or `0` if unlimited.
        video_quality_mode : ``VideoQualityMode``, `int`, Optional (Keyword only)
            The video quality of the voice channel.
        
        Returns
        -------
        instance<type<self>>
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # nsfw
        if nsfw is ...:
            nsfw = self.nsfw
        else:
            nsfw = validate_nsfw(nsfw)
        
        # status
        if status is ...:
            status = self.status
        else:
            status = validate_status(status)
        
        # video_quality_mode
        if video_quality_mode is ...:
            video_quality_mode = self.video_quality_mode
        else:
            video_quality_mode = validate_video_quality_mode(video_quality_mode)
        
        # Construct
        new = ChannelMetadataGuildVoiceBase.copy_with(
            self,
            bitrate = bitrate,
            name = name,
            parent_id = parent_id,
            permission_overwrites = permission_overwrites,
            position = position,
            region = region,
            user_limit = user_limit,
        )
        new.nsfw = nsfw
        new.status = status
        new.video_quality_mode = video_quality_mode
        return new
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            bitrate = keyword_parameters.pop('bitrate', ...),
            name = keyword_parameters.pop('name', ...),
            nsfw = keyword_parameters.pop('nsfw', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
            region = keyword_parameters.pop('region', ...),
            status = keyword_parameters.pop('status', ...),
            user_limit = keyword_parameters.pop('user_limit', ...),
            video_quality_mode = keyword_parameters.pop('video_quality_mode', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildVoiceBase._update_attributes(self, data)
        
        # nsfw
        self.nsfw = parse_nsfw(data)
        
        # video_quality_mode
        self.video_quality_mode = parse_video_quality_mode(data)
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildVoiceBase._difference_update_attributes(self, data)
        
        # nsfw
        nsfw = parse_nsfw(data)
        if self.nsfw != nsfw:
            old_attributes['nsfw'] = self.nsfw
            self.nsfw = nsfw
        
        # video_quality_mode
        video_quality_mode = parse_video_quality_mode(data)
        if self.video_quality_mode is not video_quality_mode:
            old_attributes['video_quality_mode'] = self.video_quality_mode
            self.video_quality_mode = video_quality_mode
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._update_status)
    def _update_status(self, data):
        
        # status
        self.status = parse_status(data)
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._difference_update_status)
    def _difference_update_status(self, data):
        old_attributes = {}

        # status
        status = parse_status(data)
        if self.status != status:
            old_attributes['status'] = self.status
            self.status = status
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildVoiceBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # nsfw
        put_nsfw(self.nsfw, data, defaults)
        
        # status
        put_status(self.status, data, defaults)
        
        # video_quality_mode
        put_video_quality_mode(self.video_quality_mode, data, defaults)
        
        return data
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        result = self._get_base_permissions_for(channel_entity, user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # voice channels don't have stage channel permissions
        result &= PERMISSION_DENIED_FOR_GUILD_VOICE
        
        if not result & PERMISSION_MASK_CONNECT:
            result &= PERMISSION_VOICE_DENY_CONNECTION
        
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildVoiceBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        result = self._get_base_permissions_for_roles(channel_entity, roles)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # voice channels don't have stage channel permissions
        result &= PERMISSION_DENIED_FOR_GUILD_VOICE
        
        if not result & PERMISSION_MASK_CONNECT:
            result &= PERMISSION_VOICE_DENY_CONNECTION
        
        return Permission(result)



__all__ = ('WebhookBase',)

from scarletio import copy_docs, include

from ...bases import PlaceHolder
from ...core import CHANNELS
from ...permission.permission import PERMISSION_MASK_USE_EXTERNAL_EMOJIS
from ...user import UserBase, ZEROUSER
from ...user.user.fields import validate_webhook_name

from .fields import validate_channel_id, validate_type
from .preinstanced import WebhookType


ChannelType = include('ChannelType')
create_partial_channel_from_id = include('create_partial_channel_from_id')


class WebhookBase(UserBase):
    """
    Base class for webhook like-types.
    
    Attributes
    ----------
    avatar_hash : `int`
        The webhook's avatar's hash in `uint128`.
    avatar_type : ``IconType``
        The webhook's avatar's type.
    channel_id : `int`
        The channel's identifier, where the webhook is going to send it's messages.
    id : `int`
        The webhook's unique identifier number.
    name : str
        The webhook's username.
    type : ``WebhookType``
        The webhook's type.
    """
    __slots__ = ('channel_id', 'type',)
    
    def __new__(
        cls,
        *,
        avatar = ...,
        channel_id = ...,
        name = ...,
        webhook_type = ...
    ):
        """
        Creates a new partial webhook base with the given fields.
        
        Parameters
        ----------
        channel_id : `int`, Optional (Keyword only)
            The channel's identifier, where the webhook is going to send it's messages.
        avatar : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar.
        name : `str`, Optional (Keyword only)
            The user's name.
        webhook_type : ``WebhookType``, `int`, Optional (Keyword only)
            The webhook's type.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_id
        if channel_id is ...:
            channel_id = 0
        else:
            channel_id = validate_channel_id(channel_id)
        
        # name | Do not pass `name` down since webhooks allow longer names.
        if name is ...:
            name = ''
        else:
            name = validate_webhook_name(name)
        
        # webhook_type
        if webhook_type is ...:
            webhook_type = WebhookType.none
        else:
            webhook_type = validate_type(webhook_type)
        
        # Construct
        self = UserBase.__new__(
            cls,
            avatar = avatar,
            name = ...,
        )
        self.channel_id = channel_id
        self.name = name
        self.type = webhook_type
        return self
    
    
    @copy_docs(UserBase._set_default_attributes)
    def _set_default_attributes(self):
        UserBase._set_default_attributes(self)
        
        self.channel_id = 0
        self.type = WebhookType.none
    
    
    @copy_docs(UserBase.copy)
    def copy(self):
        new = UserBase.copy(self)
        new.channel_id = self.channel_id
        new.type = self.type
        return new
    
    
    def copy_with(
        self,
        *,
        avatar = ...,
        channel_id = ...,
        name = ...,
        webhook_type = ...,
    ):
        """
        Copies the webhook base with the given fields.
        
        Parameters
        ----------
        channel_id : `int`, Optional (Keyword only)
            The channel's identifier, where the webhook is going to send it's messages.
        avatar : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar.
        name : `str`, Optional (Keyword only)
            The user's name.
        webhook_type : ``WebhookType``, `int`, Optional (Keyword only)
            The webhook's type.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_id
        if channel_id is ...:
            channel_id = self.channel_id
        else:
            channel_id = validate_channel_id(channel_id)
        
        # name | Do not pass `name` down since webhooks allow longer names.
        if name is ...:
            name = self.name
        else:
            name = validate_webhook_name(name)
        
        # webhook_type
        if webhook_type is ...:
            webhook_type = self.type
        else:
            webhook_type = validate_type(webhook_type)
        
        # Construct
        new = UserBase.copy_with(
            self,
            avatar = avatar,
            name = ...,
        )
        new.channel_id = channel_id
        new.name = name
        new.type = webhook_type
        return new
    
    
    @copy_docs(UserBase._get_hash_partial)
    def _get_hash_partial(self):
        hash_value = UserBase._get_hash_partial(self)
        
        # channel_id
        hash_value ^= self.channel_id
        
        # type
        hash_value ^= hash(self.type)
        
        return hash_value
    
    
    @copy_docs(UserBase._compare_attributes)
    def _compare_attributes(self, other):
        if not self._compare_user_attributes_extended(other):
            return False
        
        if isinstance(other, WebhookBase):
            return self._compare_webhook_attributes(other)
        
        return True
    
    
    def _compare_webhook_attributes(self, other):
        """
        Compares the two user's webhook attributes (excluding id obviously).
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other user.
        
        Returns
        -------
        is_equal : `bool`
        """
        if self.application_id != other.application_id:
            return False
        
        if self.channel_id != other.channel_id:
            return False
        
        if self.source_channel != other.source_channel:
            return False
        
        if self.source_guild != other.source_guild:
            return False
        
        if self.token != other.token:
            return False
        
        if self.type is not other.type:
            return False
        
        if self.user != other.user:
            return False
        
        return True
    
    # Placeholder
    
    application_id = PlaceHolder(
        0,
        """
        Returns the application's identifier that created the webhook.
        
        Returns
        -------
        application_id : `int`
        """
    )
    
    
    source_channel = PlaceHolder(
        None,
        """
        Returns the representation of the webhook's source channel.
        
        Returns
        -------
        source_channel : `None`, ``WebhookSourceChannel``
        """
    )
    
    
    source_guild = PlaceHolder(
        None,
        """
        Returns the representation of the webhook's source guild.
        
        Returns
        -------
        source_guild : `None`, ``WebhookSourceGuild``
        """
    )
    
    
    token = PlaceHolder(
        '',
        """
        Returns the webhook's token.
        
        Returns
        -------
        token : `str`
        """
    )
    
    
    user = PlaceHolder(
        ZEROUSER,
        """
        Returns the creator of the webhook.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
    )
    
    # Extra
    
    @property
    def bot(self):
        """
        Webhooks are always bots.
        
        Returns
        -------
        bot : `bool`
        """
        return True
    
    
    @property
    def partial(self):
        """
        Returns whether the webhook is partial.
        
        A webhook is partial, if it's respective guild is unknown.
        
        Returns
        -------
        partial : `bool`
        """
        try:
            channel = CHANNELS[self.channel_id]
        except KeyError:
            return True
        
        if channel.guild is None:
            return True
        
        return False
    
    
    @property
    def channel(self):
        """
        Returns the webhook's channel if applicable.
        
        Returns
        -------
        channel : ``Channel``
        """
        return create_partial_channel_from_id(self.channel_id, ChannelType.unknown, 0)
    
    
    @property
    def guild(self):
        """
        Returns the webhook's guild if applicable.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        try:
            channel = CHANNELS[self.channel_id]
        except KeyError:
            pass
        else:
            return channel.guild
    
    
    @copy_docs(UserBase.can_use_emoji)
    def can_use_emoji(self, emoji):
        if emoji.is_unicode_emoji():
            return True
        
        role_ids = emoji.role_ids
        if (role_ids is not None):
            return False
        
        guild = self.guild
        if guild is None:
            return False
        
        default_role = guild.default_role
        if (default_role is None):
            return False
        
        if (default_role & PERMISSION_MASK_USE_EXTERNAL_EMOJIS):
            return True
        
        return False



__all__ = ('AutoModerationActionMetadataBase',)

from scarletio import RichAttributeErrorBaseType

from ...bases import PlaceHolder
from ...channel import ChannelType, create_partial_channel_from_id


class AutoModerationActionMetadataBase(RichAttributeErrorBaseType):
    """
    Base class for ``AutoModerationAction``'s action metadata.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new action metadata instance.
        """
        return object.__new__(cls)
    
    
    def __repr__(self):
        """Returns the action metadata's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new auto moderation action metadata instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
           Auto moderation action metadata payload.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the action metadata to json serializable object.
        
        Parameters
        ----------
        defaults : `bool`
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        return {}
    
    
    def __eq__(self, other):
        """Returns whether the two action metadatas are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return True
    
    
    def __hash__(self):
        """Returns the action metadata's hash value."""
        return 0
    
    
    def copy(self):
        """
        Copies the action metadata.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return object.__new__(type(self))
    
    
    def copy_with(self):
        """
        Copies the action metadata and modifies it's attributes by the given values.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return self.copy()
    
    # ---- Place holders ----
    
    channel_id = PlaceHolder(
        0,
        """
        The channel's identifier where the alert messages are sent.
        
        Returns
        -------
        channel_id : `int`
        """
    )
    
    duration = PlaceHolder(
        0,
        """
        The timeout's duration applied on trigger.
        
        Returns
        -------
        duration : `int`
        """
    )
    
    custom_message = PlaceHolder(
        None,
        """
        A custom message that can be used as an explanation if a message is blocked.
        
        Returns
        -------
        custom_message : `None`, `str`
        """
    )
    
    # ---- Additional utility ---
    
    @property
    def channel(self):
        """
        Returns the channels where the alert messages are sent.
        
        Returns
        -------
        channel : `None`, ``Channel``
        """
        channel_id = self.channel_id
        if channel_id:
            return create_partial_channel_from_id(channel_id, ChannelType.unknown, 0)



__all__ = ()

from ..localization import Locale


def maybe_locale(value):
    """
    Converts the given value to locale if it is not already.
    
    Parameters
    ----------
    value : ``Locale``, `str`
        Value to convert to locale.
    
    Returns
    -------
    locale : ``Locale``
    
    Raises
    ------
    TypeError
        - `value`'s type is neither ``Locale`` nor `str`.
    """
    if isinstance(value, Locale):
        locale = value
    
    elif isinstance(value, str):
        locale = Locale(value)
    
    else:
        raise TypeError(
            f'`locale`-s can be `{Locale.__name}`, `str`, got {type(value).__name__}; {value!r}.'
        )
    
    return locale


def with_translation(source_value, localized_dictionary, translation_table, replace):
    """
    Applies translation of for the given value to the given localized dictionary from the given translation table.
    
    Parameters
    ----------
    source_value : `None`, `str`
        The value to get translations for.
    localized_dictionary : `None`, `dict` of ((`str`, ``Locale``), `str`) items
        Localized dictionary to apply the translations to.
    translation_table : `dict` of ((``Locale``, `str`), (`None`, `dict` (`str`, (`None`, `str`)) items)) items
        Translation table to pull localizations from.
    replace : `bool`
        Whether actual translation should be replaced.
    
    Returns
    -------
    localized_dictionary : `None`, `dict` of ((`str`, ``Locale``), `str`) items
        New localized dictionary.
    """
    if (source_value is not None):
        for locale, relations in translation_table.items():
            if relations is None:
                continue
            
            try:
                translation = relations[source_value]
            except KeyError:
                continue
            
            if translation is None:
                continue
            
            locale = maybe_locale(locale)
            
            if localized_dictionary is None:
                localized_dictionary = {}
            
            if replace:
                localized_dictionary[locale] = translation
            else:
                localized_dictionary.setdefault(locale, translation)
    
    return localized_dictionary



from .constants import *
from .fields import *
from .soundboard_sound import *
from .utils import *


__all__ = (
    *constants.__all__,
    *fields.__all__,
    *soundboard_sound.__all__,
    *utils.__all__,
)



__all__ = ('EXTRA_EMBED_TYPES',)

from .preinstanced import EmbedType


TITLE_LENGTH_MAX = 256
DESCRIPTION_LENGTH_MAX = 4096
FIELDS_LENGTH_MAX = 25
URL_LENGTH_MAX = 2048


EXTRA_EMBED_TYPES = frozenset((
    EmbedType.application_news,
    EmbedType.article,
    EmbedType.auto_moderation_message,
    EmbedType.gifv,
    EmbedType.image,
    EmbedType.link,
    EmbedType.tweet,
    EmbedType.video,
))



from .conversions import *

from .builder_base import *
from .constants import *
from .conversion import *
from .descriptor import *
from .serialization_configuration import *


__all__ = (
    *conversions.__all__,
    
    *builder_base.__all__,
    *constants.__all__,
    *conversion.__all__,
    *descriptor.__all__,
    *serialization_configuration.__all__,
)



__all__ = ('WebhookType',)

from ...bases import Preinstance as P, PreinstancedBase


class WebhookType(PreinstancedBase, value_type = int):
    """
    Represents a webhook's type.
    
    Attributes
    ----------
    name : `str`
        The name of the webhook type.
    
    value : `int`
        The discord side identifier value of the webhook type.
    
    Type Attributes
    ---------------
    Every predefined webhook type can be accessed as type attribute as well:
    
    +-----------------------+---------------+-------+
    | Type attribute name   | name          | value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | bot                   | bot           | 1     |
    +-----------------------+---------------+-------+
    | server                | server        | 2     |
    +-----------------------+---------------+-------+
    | application           | application   | 3     |
    +-----------------------+---------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    bot = P(1, 'bot')
    server = P(2, 'server')
    application = P(3, 'application')
